# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0384: OMEMO Encryption

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Literal

from collections.abc import Generator
from pathlib import Path

import nbxmpp
from nbxmpp.errors import is_error
from nbxmpp.modules.omemo import create_omemo_message
from nbxmpp.modules.omemo import get_key_transport_message
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import EncryptionData
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import OMEMOMessage
from nbxmpp.structs import StanzaHandler
from omemo_dr.const import OMEMOTrust
from omemo_dr.exceptions import DecryptionFailed
from omemo_dr.exceptions import DuplicateMessage
from omemo_dr.exceptions import KeyExchangeMessage
from omemo_dr.exceptions import MessageNotForDevice
from omemo_dr.exceptions import SelfMessage
from omemo_dr.identitykey import IdentityKey
from omemo_dr.session_manager import OMEMOSessionManager
from omemo_dr.structs import OMEMOBundle
from omemo_dr.structs import OMEMOConfig

from gajim.common import app
from gajim.common import configpaths
from gajim.common import ged
from gajim.common import types
from gajim.common.const import EncryptionInfoMsg
from gajim.common.const import Trust as GajimTrust
from gajim.common.const import XmppUriQuery
from gajim.common.events import EncryptionInfo
from gajim.common.events import MucAdded
from gajim.common.events import MucDiscoUpdate
from gajim.common.events import SignedIn
from gajim.common.i18n import _
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.util import as_task
from gajim.common.modules.util import event_node
from gajim.common.modules.util import prepare_stanza
from gajim.common.storage.omemo import OMEMOStorage
from gajim.common.structs import OutgoingMessage
from gajim.common.util.decorators import cache_with_ttl
from gajim.common.util.decorators import event_filter

ALLOWED_TAGS = [
    ('request', Namespace.RECEIPTS),
    ('active', Namespace.CHATSTATES),
    ('gone', Namespace.CHATSTATES),
    ('inactive', Namespace.CHATSTATES),
    ('paused', Namespace.CHATSTATES),
    ('composing', Namespace.CHATSTATES),
    ('markable', Namespace.CHATMARKERS),
    ('no-store', Namespace.HINTS),
    ('store', Namespace.HINTS),
    ('no-copy', Namespace.HINTS),
    ('no-permanent-store', Namespace.HINTS),
    ('replace', Namespace.CORRECT),
    ('thread', None),
    ('reply', Namespace.REPLY),
    ('fallback', Namespace.FALLBACK),
    ('origin-id', Namespace.SID),
    ('reactions', Namespace.REACTIONS),
]

DeviceIdT = int
IdentityT = tuple[DeviceIdT, IdentityKey]

class OMEMO(BaseModule):

    _nbxmpp_extends = 'OMEMO'
    _nbxmpp_methods = [
        'set_devicelist',
        'request_devicelist',
        'set_bundle',
        'request_bundle',
    ]

    def __init__(self, client: types.Client) -> None:
        BaseModule.__init__(self, client)

        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._message_received,
                          ns=Namespace.OMEMO_TEMP,
                          priority=9),
        ]

        self._register_pubsub_handler(self._devicelist_notification_received)
        self.register_events([
            ('signed-in', ged.CORE, self._on_signed_in),
            ('muc-disco-update', ged.PREGUI, self._on_muc_disco_update),
            ('muc-added', ged.PREGUI, self._on_muc_added)
        ])

        self.allow_groupchat = True

        self._own_jid = self._client.get_own_jid().bare

        data_dir = Path(configpaths.get('MY_DATA'))
        db_path = data_dir / f'omemo_{self._own_jid}.db'
        storage = OMEMOStorage(self._account, db_path, self._log)

        omemo_config = OMEMOConfig(default_prekey_amount=100,
                                   min_prekey_amount=80,
                                   spk_archive_seconds=86400 * 15,
                                   spk_cycle_seconds=86400 * 30,
                                   unacknowledged_count=2000)

        self._backend = OMEMOSessionManager(
            self._own_jid, storage, omemo_config, self._account)
        self._backend.register_signal('republish-bundle',
                                      self._on_republish_bundle)

        self._omemo_groupchats: set[str] = set()
        self._muc_temp_store: dict[bytes, str] = {}

    @event_filter(['account'])
    def _on_signed_in(self, _event: SignedIn) -> None:
        self._log.info('Publish our bundle after sign in')
        self.set_bundle()
        self.request_devicelist()

    @event_filter(['account'])
    def _on_muc_disco_update(self, event: MucDiscoUpdate) -> None:
        before = self._is_omemo_groupchat(event.jid)
        self._check_if_omemo_capable(event.jid)
        after = self._is_omemo_groupchat(event.jid)

        if not before and after:
            self._set_group_members(event.jid)

    @event_filter(['account'])
    def _on_muc_added(self, event: MucAdded) -> None:
        client = app.get_client(event.account)
        contact = client.get_module('Contacts').get_contact(event.jid)
        if not isinstance(contact, GroupchatContact):
            self._log.warning('%s is not a groupchat contact', contact)
            return

        # Event is triggert on every join, avoid multiple connects
        contact.disconnect_all_from_obj(self)
        contact.connect('room-joined', self._on_room_joined)
        contact.connect('user-affiliation-changed', self._on_affiliation_change)
        contact.connect('room-affiliation-changed', self._on_affiliation_change)
        contact.connect('room-affiliations-received', self._on_affiliation_change)

    def _on_room_joined(self,
                        contact: GroupchatContact,
                        _signal_name: str
                        ) -> None:

        self._check_if_omemo_capable(contact.jid)

    def _on_affiliation_change(
        self,
        contact: GroupchatContact,
        _signal_name: str,
        *args: Any,
    ) -> None:

        if not self._is_omemo_groupchat(contact.jid):
            return

        self._set_group_members(contact.jid)

    def _set_group_members(self, muc_jid: JID) -> None:
        self._log.info("Set group members: %s", muc_jid)
        affiliations = self._client.get_module('MUC').get_affiliations(muc_jid)
        members: set[JID] = set()
        members.update(*affiliations.values())
        str_jids = set(map(str, members))

        self._backend.set_group_members(str(muc_jid), str_jids)

        for jid in str_jids:
            if not self._is_contact_in_roster(jid):
                self._log.info('%s not in roster, query devicelist...', jid)
                self._request_device_list_ttl(jid)

    def _on_republish_bundle(self,
                             _session: OMEMOSessionManager,
                             _signal_name: str,
                             bundle: OMEMOBundle
                             ) -> None:

        self.set_bundle(bundle=bundle)

    @property
    def backend(self) -> OMEMOSessionManager:
        return self._backend

    def check_send_preconditions(self, contact: types.ChatContactT) -> bool:
        jid = str(contact.jid)
        if isinstance(contact, GroupchatContact):
            if not self._groupchat_pre_conditions_satisfied(contact):
                return False

        else:
            if not self._chat_pre_conditions_satisfied(contact):
                return False

        if self.backend.get_identity_infos(jid,
                                           only_active=True,
                                           trust=OMEMOTrust.UNDECIDED):
            self._log.info('Undecided keys for %s', jid)
            app.ged.raise_event(EncryptionInfo(
                account=contact.account,
                jid=contact.jid,
                type=EncryptionInfoMsg.UNDECIDED_FINGERPRINTS,
                message=EncryptionInfoMsg.UNDECIDED_FINGERPRINTS.value,
            ))
            return False

        self._log.debug('Sending message to %s', jid)

        return True

    def _groupchat_pre_conditions_satisfied(
            self,
            contact: types.GroupchatContactT
        ) -> bool:

        if not self._is_omemo_groupchat(contact.jid):
            app.ged.raise_event(EncryptionInfo(
                account=contact.account,
                jid=contact.jid,
                type=EncryptionInfoMsg.BAD_OMEMO_CONFIG,
                message=EncryptionInfoMsg.BAD_OMEMO_CONFIG.value,
            ))
            return False

        jid = str(contact.jid)
        has_trusted_keys = False
        for member_jid in self.backend.get_group_members(jid, without_self=False):
            self._request_bundles_for_new_devices(member_jid)
            if self._has_trusted_keys(member_jid):
                has_trusted_keys = True

        if not has_trusted_keys:
            self._log.info('No trusted keys for %s', jid)
            app.ged.raise_event(EncryptionInfo(
                account=contact.account,
                jid=contact.jid,
                type=EncryptionInfoMsg.NO_FINGERPRINTS,
                message=EncryptionInfoMsg.NO_FINGERPRINTS.value,
            ))
            return False
        return True

    def _chat_pre_conditions_satisfied(
            self,
            contact: types.ChatContactT
        ) -> bool:

        jid = str(contact.jid)
        if not self.backend.get_devices(jid, without_self=True):
            self.request_devicelist(jid)
            app.ged.raise_event(EncryptionInfo(
                account=contact.account,
                jid=contact.jid,
                type=EncryptionInfoMsg.QUERY_DEVICES,
                message=EncryptionInfoMsg.QUERY_DEVICES.value,
            ))
            return False

        if not self._has_trusted_keys(jid):
            self._log.info('No trusted keys for %s', jid)
            app.ged.raise_event(EncryptionInfo(
                account=contact.account,
                jid=contact.jid,
                type=EncryptionInfoMsg.NO_FINGERPRINTS,
                message=EncryptionInfoMsg.NO_FINGERPRINTS.value,
            ))
            return False
        return True

    def _is_omemo_groupchat(self, room_jid: JID) -> bool:
        return str(room_jid) in self._omemo_groupchats

    def encrypt_message(self, message: OutgoingMessage) -> bool:
        if not message.has_text():
            return False

        remote_jid = message.contact.jid

        client = app.get_client(self._account)
        contact = client.get_module('Contacts').get_contact(remote_jid)

        text = message.get_text()
        assert text is not None
        omemo_message = self.backend.encrypt(str(remote_jid),
                                             text,
                                             groupchat=contact.is_groupchat)
        if omemo_message is None:
            raise Exception('Encryption error')

        create_omemo_message(message.get_stanza(), omemo_message,
                             node_whitelist=ALLOWED_TAGS)

        if message.is_groupchat:
            self._muc_temp_store[omemo_message.payload] = text

        message.set_encryption(
            EncryptionData(
                protocol='OMEMO',
                key='Unknown',
                trust=GajimTrust[OMEMOTrust.VERIFIED.name]
            )
        )

        self._debug_print_stanza(message.get_stanza())
        return True

    def _send_key_transport_message(self,
                                    typ: Literal['chat', 'groupchat'],
                                    jid: str,
                                    devices: list[int]
                                    ) -> None:

        omemo_message = self.backend.encrypt_key_transport(jid, devices)
        if omemo_message is None:
            self._log.warning('Key transport message to %s (%s) failed',
                              jid, devices)
            return

        transport_message = get_key_transport_message(typ, jid, omemo_message)
        self._log.info('Send key transport message to %s (%s)', jid, devices)
        self._client.send_stanza(transport_message)

    def _message_received(self,
                          _client: types.NBXMPPClient,
                          stanza: Message,
                          properties: MessageProperties
                          ) -> None:

        if properties.omemo is None:
            return

        if properties.carbon is not None and properties.carbon.is_sent:
            from_jid = self._own_jid

        elif properties.is_mam_message:
            from_jid = self._process_mam_message(properties)

        elif properties.from_muc:
            from_jid = self._process_muc_message(properties)

        else:
            from_jid = properties.jid.bare

        if from_jid is None:
            return

        self._log.info('Message received from: %s', from_jid)

        assert isinstance(properties.omemo, OMEMOMessage)
        try:
            plaintext, fingerprint, trust = self.backend.decrypt_message(
                properties.omemo, from_jid)
        except (KeyExchangeMessage, DuplicateMessage):
            raise NodeProcessed

        except SelfMessage:
            if not properties.from_muc:
                raise NodeProcessed

            if properties.omemo.payload not in self._muc_temp_store:
                self._log.warning("Can't decrypt own group chat message")
                return

            del self._muc_temp_store[properties.omemo.payload]
            return

        except DecryptionFailed:
            return

        except MessageNotForDevice:
            if properties.omemo.payload is None:
                # Key Transport message for another device
                return

            plaintext = _('This message was encrypted with OMEMO, '
                          'but not for your device.')
            # Neither trust nor fingerprint can be verified if we didn't
            # successfully decrypt the message
            trust = OMEMOTrust.UNTRUSTED
            fingerprint = None

        prepare_stanza(stanza, plaintext)
        self._debug_print_stanza(stanza)
        properties.encrypted = EncryptionData(
            protocol='OMEMO',
            key=fingerprint or 'Unknown',
            trust=GajimTrust[trust.name]
        )

    def _process_muc_message(self,
                             properties: MessageProperties
                             ) -> str | None:

        assert properties.jid is not None
        resource = properties.jid.resource
        if properties.muc_ofrom is not None:
            # History Message from MUC
            return properties.muc_ofrom.bare

        contact = self._client.get_module('Contacts').get_contact(
            properties.jid)
        assert isinstance(contact, GroupchatParticipant)
        if contact.real_jid is not None:
            return contact.real_jid.bare

        self._log.error(
            'Unable to find jid for group chat message from %s', resource)
        return None

    def _process_mam_message(self,
                             properties: MessageProperties
                             ) -> str | None:

        self._log.info('Message received, archive: %s', properties.mam.archive)
        if properties.from_muc:
            self._log.info('MUC MAM Message received')
            if properties.muc_user is None or properties.muc_user.jid is None:
                self._log.warning('Received MAM message which can '
                                  'not be mapped to a real jid')
                return None
            return properties.muc_user.jid.bare
        return properties.from_.bare

    def _is_contact_in_roster(self, jid: str) -> bool:
        if jid == self._own_jid:
            return True

        roster_item = self._client.get_module('Roster').get_item(jid)
        if roster_item is None:
            return False

        contact = self._client.get_module('Contacts').get_contact(jid)
        assert isinstance(contact, BareContact)
        return contact.subscription == 'both'

    def _check_if_omemo_capable(self, jid: JID) -> None:
        disco_info = app.storage.cache.get_last_disco_info(jid)
        assert disco_info is not None
        if disco_info.muc_is_members_only and disco_info.muc_is_nonanonymous:
            self._log.info('OMEMO room discovered: %s', jid)
            self._omemo_groupchats.add(str(jid))
        else:
            self._log.info('OMEMO room removed due to config change: %s', jid)
            self._omemo_groupchats.discard(str(jid))

    def _request_bundles_for_new_devices(self, jid_: str) -> None:
        for jid in [jid_, self._own_jid]:
            device_ids = self.backend.get_devices_without_sessions(jid)
            for device_id in device_ids:
                self._request_bundle_ttl(jid, device_id)

    def _has_trusted_keys(self, jid: str) -> bool:
        infos = self.backend.get_identity_infos(
                jid,
                only_active=True,
                trust=[OMEMOTrust.VERIFIED, OMEMOTrust.BLIND])
        return bool(infos)

    def set_bundle(self, bundle: OMEMOBundle | None = None) -> None:
        if bundle is None:
            bundle = self.backend.get_bundle(Namespace.OMEMO_TEMP)
        self._nbxmpp('OMEMO').set_bundle(bundle,
                                         self.backend.get_our_device())

    @cache_with_ttl(ttl=7200)
    def _request_bundle_ttl(self, jid: str, device_id: int) -> None:
        self.request_bundle(jid, device_id)

    @as_task
    def request_bundle(self, jid: str, device_id: int) -> Generator[Any, Any]:
        _task = yield  # noqa: F841

        self._log.info('Request device bundle %s %s', device_id, jid)

        bundle = yield self._nbxmpp('OMEMO').request_bundle(jid, device_id)

        if is_error(bundle) or bundle is None:
            self._log.info('Bundle request failed: %s %s: %s',
                           jid, device_id, bundle)
            return

        try:
            self.backend.build_session(jid, bundle)
        except Exception as error:
            self._log.error('Building session failed: %s', error)
            return

        self._log.info('Session created for: %s', jid)
        # TODO: In MUC we should send a groupchat message
        self._send_key_transport_message('chat', jid, [device_id])

        # Trigger dialog to trust new Fingerprints if
        # the Chat Window is Open
        app.ged.raise_event(EncryptionInfo(
            account=self._account,
            jid=JID.from_string(jid),
            type=EncryptionInfoMsg.SESSION_BUILD,
            message=EncryptionInfoMsg.SESSION_BUILD.value.format(device_id=device_id),
        ))

    def set_devicelist(self, devicelist: list[int] | None = None) -> None:
        devicelist_: set[int] = {self.backend.get_our_device()}
        if devicelist is not None:
            devicelist_.update(devicelist)
        self._log.info('Publishing own devicelist: %s', devicelist_)
        self._nbxmpp('OMEMO').set_devicelist(devicelist_)

    def clear_devicelist(self) -> None:
        self.backend.update_devicelist(
            self._own_jid, [self.backend.get_our_device()])
        self.set_devicelist()

    @cache_with_ttl(ttl=7200)
    def _request_device_list_ttl(self, jid: str) -> None:
        self.request_devicelist(jid)

    @as_task
    def request_devicelist(self, jid: str | None = None) -> Generator[Any, Any]:
        _task = yield  # noqa: F841

        if jid is None:
            jid = self._own_jid

        self._log.info('Request devicelist for %s', jid)

        devicelist = yield self._nbxmpp('OMEMO').request_devicelist(jid=jid)
        if is_error(devicelist) or devicelist is None:
            self._log.info('Devicelist request failed: %s %s', jid, devicelist)
            devicelist = []

        self._process_devicelist_update(jid, cast(list[int], devicelist))

    @event_node(Namespace.OMEMO_TEMP_DL)
    def _devicelist_notification_received(self,
                                          _client: types.NBXMPPClient,
                                          _stanza: Message,
                                          properties: MessageProperties
                                          ) -> None:

        assert properties.pubsub_event is not None
        if properties.pubsub_event.retracted:
            return

        devicelist = cast(list[int], properties.pubsub_event.data or [])

        self._process_devicelist_update(str(properties.jid), devicelist)

    def _process_devicelist_update(self,
                                   jid: str,
                                   devicelist: list[int]
                                   ) -> None:

        own_devices = jid is None or self._client.get_own_jid().bare_match(jid)
        if own_devices:
            jid = self._own_jid

        self._log.info('Received device list for %s: %s', jid, devicelist)
        # Pass a copy, we need the full list for potential set_devicelist()
        self.backend.update_devicelist(jid, list(devicelist))

        if own_devices:
            if not self.backend.is_our_device_published():
                # Our own device_id is not in the list, it could be
                # overwritten by some other client
                self.set_devicelist(devicelist)

        self._request_bundles_for_new_devices(jid)

    def _debug_print_stanza(self, stanza: nbxmpp.Node) -> None:
        stanzastr = '\n' + stanza.__str__(fancy=True)
        stanzastr = stanzastr[0:-1]
        self._log.debug(stanzastr)

    def compose_trust_uri(self, jid: JID) -> str:
        verified_identities = [
            (info.device_id, info.public_key)
            for info in self._backend.get_identity_infos(
                jid.bare, only_active=True, trust=OMEMOTrust.VERIFIED)
        ]
        if self._client.is_own_jid(jid):
            verified_identities.insert(0, self._backend.get_our_identity())
        return compose_trust_uri(jid, verified_identities)

    def cleanup(self) -> None:
        BaseModule.cleanup(self)
        self._backend.destroy()
        del self._backend


def compose_trust_uri(jid: JID, devices: list[IdentityT]) -> str:
    query = (
        XmppUriQuery.MESSAGE.value,
        [(f'omemo-sid-{sid}', ik.get_fingerprint()) for sid, ik in devices]
    ) if devices else None
    uri = jid.new_as_bare().to_iri(query)
    return uri



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Final

import locale
import logging

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.i18n import _
from gajim.common.i18n import get_default_lang
from gajim.common.i18n import get_short_lang_code

from gajim.gtk.completion.base import BaseCompletionListItem
from gajim.gtk.completion.base import BaseCompletionProvider
from gajim.gtk.completion.base import BaseCompletionViewItem
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string

log = logging.getLogger("gajim.gtk.completion.emoji")

EMOJI_DATA_ENTRY_T = tuple[list[int], str, str, list[str], list[str], int]
MAX_COMPLETION_ENTRIES = 8
FALLBACK_LOCALE = "en"

SKIN_TONE_MODIFIERS = (0x1F3FB, 0x1F3FC, 0x1F3FD, 0x1F3FE, 0x1F3FF)


def get_locale_fallbacks(desired: str) -> list[str]:
    """
    Returns full list of locales to try loading emoji data in, in the order of
    decreasing preference and specificity.  E.g., ['de', 'en']
    for desired == 'de'.
    """
    result = [desired]
    if FALLBACK_LOCALE not in result:
        result.append(FALLBACK_LOCALE)

    return result


def generate_unicode_sequence(c_sequence: list[int]) -> tuple[bool, str]:
    """
    Generates a unicode sequence from a list of codepoints
    """
    has_skin_tone_variation = False
    u_sequence = ""
    for codepoint in c_sequence:
        if codepoint == 0:
            has_skin_tone_variation = True
            codepoint = 0xFE0F

        if codepoint == 0x1F3FB:
            has_skin_tone_variation = True
            continue

        u_sequence += chr(codepoint)
    return has_skin_tone_variation, u_sequence


def generate_skin_tone_sequence(c_sequence: list[int], modifier: int) -> str:
    """
    Replaces GTKs placeholder '0' for skin tone modifiers
    with a given modifier
    """
    u_sequence = ""
    for codepoint in c_sequence:
        if codepoint in (0, 0x1F3FB):
            codepoint = modifier
        u_sequence += chr(codepoint)
    return u_sequence


def try_load_raw_emoji_data(locale: str) -> GLib.Bytes | None:
    # Sources of emoji data can be found at:
    # https://gitlab.gnome.org/GNOME/gtk/-/tree/main/gtk/emoji
    emoji_data_resource = f"/org/gtk/libgtk/emoji/{locale}.data"

    # some distribution do not register locale emoji resources, so let's do it
    try:
        res = Gio.resource_load(f"/usr/share/gtk-4.0/emoji/{locale}.gresource")
    except GLib.Error:
        pass
    else:
        Gio.resources_register(res)

    try:
        bytes_ = Gio.resources_lookup_data(
            emoji_data_resource, Gio.ResourceLookupFlags.NONE
        )
        assert bytes_ is not None
        log.info("Loaded emoji data resource for locale %s", locale)
        return bytes_
    except GLib.Error as error:
        log.info("Loading emoji data resource for locale %s failed: %s", locale, error)
        return None


def parse_emoji_data(
    bytes_data: GLib.Bytes, loc: str
) -> Gio.ListStore[EmojiCompletionListItem]:
    variant = GLib.Variant.new_from_bytes(
        # Reference for the data format:
        # https://gitlab.gnome.org/GNOME/gtk/-/blob/main/gtk/emoji/convert-emoji.c#L25
        GLib.VariantType("a(aussasasu)"),
        bytes_data,
        True,
    )

    iterable: list[EMOJI_DATA_ENTRY_T] = variant.unpack()

    store = Gio.ListStore(item_type=EmojiCompletionListItem)
    for (
        c_sequence,
        _short_name,
        trans_short_name,
        _keywords,
        trans_keywords,
        _group,
    ) in iterable:
        # Example item:
        # (
        #     [128515],
        #     'grinning face with big eyes',
        #     'grinsendes Gesicht mit großen Augen',
        #     ['face', 'mouth', 'open', 'smile'],
        #     ['gesicht', 'grinsendes gesicht mit großen augen', 'lol', 'lustig', 'lächeln'],  # noqa: E501
        #     1
        # )
        # If '0' is in c_sequence its a placeholder for skin tone modifiers

        has_skin_variation, u_sequence = generate_unicode_sequence(c_sequence)
        keywords_string = ", ".join(trans_keywords)

        u_mod_sequences: dict[str, str] = {}
        if has_skin_variation:
            for index, modifier in enumerate(SKIN_TONE_MODIFIERS, start=1):
                u_mod_sequences[f"var{index}"] = generate_skin_tone_sequence(
                    c_sequence, modifier
                )

        item = EmojiCompletionListItem(
            emoji=u_sequence,
            short_name=trans_short_name,
            keywords=f"[ {keywords_string} ]",
            search=f"{trans_short_name}|{'|'.join(trans_keywords)}",
            has_skin_variation=has_skin_variation,
            **u_mod_sequences,
        )
        store.append(item)

    store.sort(_sort_short_name)
    return store


@staticmethod
def _sort_short_name(
    item1: EmojiCompletionListItem, item2: EmojiCompletionListItem
) -> int:
    return locale.strcoll(item1.short_name, item2.short_name)


class EmojiCompletionListItem(BaseCompletionListItem, GObject.Object):
    __gtype_name__ = "EmojiCompletionListItem"

    emoji = GObject.Property(type=str)
    short_name = GObject.Property(type=str)
    keywords = GObject.Property(type=str)
    search = GObject.Property(type=str)
    has_skin_variation = GObject.Property(type=bool, default=False)
    var1 = GObject.Property(type=str, default="")
    var2 = GObject.Property(type=str, default="")
    var3 = GObject.Property(type=str, default="")
    var4 = GObject.Property(type=str, default="")
    var5 = GObject.Property(type=str, default="")

    def get_text(self) -> str:
        return self.emoji


@Gtk.Template(string=get_ui_string("emoji_completion_view_item.ui"))
class EmojiCompletionViewItem(
    BaseCompletionViewItem[EmojiCompletionListItem], Gtk.Stack, SignalManager
):
    __gtype_name__ = "EmojiCompletionViewItem"
    css_class = "emoji-completion"

    _emoji_label: Gtk.Label = Gtk.Template.Child()
    _short_name_label: Gtk.Label = Gtk.Template.Child()
    _keywords_label: Gtk.Label = Gtk.Template.Child()
    _var1_button: Gtk.Button = Gtk.Template.Child()
    _var2_button: Gtk.Button = Gtk.Template.Child()
    _var3_button: Gtk.Button = Gtk.Template.Child()
    _var4_button: Gtk.Button = Gtk.Template.Child()
    _var5_button: Gtk.Button = Gtk.Template.Child()

    has_skin_variation = GObject.Property(type=bool, default=False)

    def __init__(self) -> None:
        super().__init__()
        Gtk.Stack.__init__(self)
        SignalManager.__init__(self)

        controller = Gtk.GestureClick(
            button=Gdk.BUTTON_SECONDARY, propagation_phase=Gtk.PropagationPhase.CAPTURE
        )
        self._connect(controller, "pressed", self._on_button_press)
        self.add_controller(controller)

        self._connect(self._var1_button, "clicked", self._on_var_button_clicked)
        self._connect(self._var2_button, "clicked", self._on_var_button_clicked)
        self._connect(self._var3_button, "clicked", self._on_var_button_clicked)
        self._connect(self._var4_button, "clicked", self._on_var_button_clicked)
        self._connect(self._var5_button, "clicked", self._on_var_button_clicked)

    def _on_button_press(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        _x: float,
        _y: float,
    ) -> None:
        if not self.props.has_skin_variation:  # pyright: ignore
            return

        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
        self.set_visible_child_name("variations")

    def _on_var_button_clicked(self, button: Gtk.Button) -> None:
        emoji = button.get_label()
        view = cast(Gtk.Widget, self.get_parent().get_parent())  # pyright: ignore
        view.emit("extended-activate", emoji)

    def bind(self, obj: EmojiCompletionListItem) -> None:
        bind_spec = [
            ("has_skin_variation", self, "has_skin_variation"),
            ("emoji", self._emoji_label, "label"),
            ("short_name", self._short_name_label, "label"),
            ("keywords", self._keywords_label, "label"),
            ("var1", self._var1_button, "label"),
            ("var2", self._var2_button, "label"),
            ("var3", self._var3_button, "label"),
            ("var4", self._var4_button, "label"),
            ("var5", self._var5_button, "label"),
        ]

        for source_prop, widget, target_prop in bind_spec:
            bind = obj.bind_property(
                source_prop, widget, target_prop, GObject.BindingFlags.SYNC_CREATE
            )
            self._bindings.append(bind)

    def unbind(self) -> None:
        self.set_visible_child_name("emoji")
        for bind in self._bindings:
            bind.unbind()
        self._bindings.clear()

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Stack.do_unroot(self)
        app.check_finalize(self)


class EmojiCompletionProvider(BaseCompletionProvider):
    trigger_char: Final = ":"
    name = _("Emojis")

    def __init__(self) -> None:
        self._load_complete = False

        expression = Gtk.PropertyExpression.new(EmojiCompletionListItem, None, "search")
        self._string_filter = Gtk.StringFilter(expression=expression)

        self._filter_model = Gtk.FilterListModel(filter=self._string_filter)
        self._model = Gtk.SliceListModel(
            model=self._filter_model, size=MAX_COMPLETION_ENTRIES
        )

    def get_model(
        self,
    ) -> tuple[Gio.ListModel[EmojiCompletionListItem], type[EmojiCompletionViewItem]]:
        return self._model, EmojiCompletionViewItem

    @staticmethod
    def _load_emoji_data() -> Gio.ListStore[EmojiCompletionListItem]:
        app_locale = get_default_lang()
        log.info("Loading emoji data; application locale is %s", app_locale)
        short_locale = get_short_lang_code(app_locale)
        locales = get_locale_fallbacks(short_locale)
        try:
            log.debug("Trying locales %s", locales)
            raw_emoji_data: GLib.Bytes | None = None
            for loc in locales:
                raw_emoji_data = try_load_raw_emoji_data(loc)
                if raw_emoji_data:
                    break
            else:
                raise RuntimeError(f"No resource could be loaded; tried {locales}")

            return parse_emoji_data(raw_emoji_data, loc)
        except Exception as err:
            log.warning("Unable to load emoji data: %s", err)
            return Gio.ListStore(item_type=EmojiCompletionListItem)

    def check(self, candidate: str, start_iter: Gtk.TextIter) -> bool:
        return candidate.startswith(self.trigger_char)

    def populate(self, candidate: str, contact: Any) -> bool:
        candidate = candidate.lstrip(self.trigger_char)
        if not candidate or not len(candidate) > 1:
            # Don't activate until a sufficient # of chars have been typed,
            # which is chosen to be > 1 to not interfere with ASCII smilies
            # consisting of a colon and 1 single other char.
            return False

        if not app.settings.get("enable_emoji_shortcodes"):
            return False

        if not self._load_complete:
            model = self._load_emoji_data()
            self._filter_model.set_model(model)
            self._load_complete = True

        self._string_filter.set_search(candidate)
        return self._model.get_n_items() > 0



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import itertools
import logging
import multiprocessing
import re
import threading
from concurrent.futures import Future
from functools import partial

from gi.repository import GLib
from gi.repository import Gtk

from gajim.common import app
from gajim.common.helpers import determine_proxy
from gajim.common.multiprocess.http import CancelledError
from gajim.common.multiprocess.url_preview import generate_url_preview
from gajim.common.open_graph_parser import OpenGraphData
from gajim.common.regex import HTTPS_URL_RX
from gajim.common.types import ChatContactT

from gajim.gtk.preview.open_graph import OpenGraphPreviewWidget
from gajim.gtk.util.misc import check_finalize
from gajim.gtk.util.misc import container_remove_all

log = logging.getLogger("gajim.gtk.message_url_previews")

MAX_URL_PREVIEWS = 3


class _MISSING:
    pass


MISSING = _MISSING()


class MessageURLPreviews(Gtk.Box):
    def __init__(self) -> None:
        Gtk.Box.__init__(
            self, spacing=12, orientation=Gtk.Orientation.VERTICAL, visible=False
        )

        mp_context = multiprocessing.get_context("spawn")
        self._manager = mp_context.Manager()

        self._preview_timeout_id = None
        self._current_previews: dict[str, OpenGraphPreviewWidget] = {}
        self._dismissed_previews: set[str] = set()

        self._requests_in_progress: dict[str, threading.Event] = {}
        self._request_cache: dict[str, OpenGraphData | None] = {}

        self._contact: ChatContactT | None = None

    def get_open_graph_data(self) -> dict[str, OpenGraphData] | None:
        if not self._current_previews:
            return None

        data: dict[str, OpenGraphData] = {}
        for url, preview in self._current_previews.items():
            if og_data := preview.get_open_graph():
                data[url] = og_data

        return data

    def switch_contact(self, contact: ChatContactT) -> None:
        self.clear()
        self._contact = contact

    def clear(self) -> None:
        if self._preview_timeout_id is not None:
            GLib.source_remove(self._preview_timeout_id)
            self._preview_timeout_id = None

        for event in self._requests_in_progress.values():
            event.set()

        self._requests_in_progress.clear()
        self._current_previews.clear()
        self._dismissed_previews.clear()

        container_remove_all(self)

        self.set_visible(False)

    def generate_url_previews(self, text: str) -> None:
        if self._contact is None:
            return

        if self._contact.is_groupchat:
            if not app.settings.get("gc_enable_link_preview_default"):
                return
        else:
            if not app.settings.get("enable_link_preview_default"):
                return

        if self._contact.settings.get("encryption"):
            return

        if self._preview_timeout_id is not None:
            GLib.source_remove(self._preview_timeout_id)

        self._preview_timeout_id = GLib.timeout_add(500, self._find_urls, text)

    def _find_urls(self, text: str) -> None:
        self._preview_timeout_id = None

        # We use lists here to preserve a stable and consistent order
        matches = itertools.islice(re.finditer(HTTPS_URL_RX, text), MAX_URL_PREVIEWS)
        urls = [match.group() for match in matches]

        # Don't trigger on unrealistic short urls
        urls = list(filter(lambda u: len(u) > 10, urls))

        # Remove dismissed urls if they are not present in the text anymore
        self._dismissed_previews = set(urls) & self._dismissed_previews

        # Ignore already dismissed urls
        for url in self._dismissed_previews:
            if url in urls:
                urls.remove(url)

        self._update_preview_list(urls)

    def _request_url_preview(self, url: str) -> None:
        log.info("Request preview for: %s", url)

        event = self._manager.Event()

        assert self._contact is not None
        proxy = determine_proxy(self._contact.account)

        try:
            future = app.process_pool.submit(
                generate_url_preview,
                url,
                event,
                None if proxy is None else proxy.get_uri(),
            )
        except Exception as error:
            log.warning("Unable to generate preview: %s %s", url, error)
            return

        future.add_done_callback(partial(GLib.idle_add, self._on_request_finished, url))

        self._requests_in_progress[url] = event

    def _on_request_finished(
        self, url: str, future: Future[OpenGraphData | None]
    ) -> None:
        log.info("Request finished for %s", url)
        self._requests_in_progress.pop(url, None)

        og_data = None
        try:
            og_data = future.result()
        except CancelledError:
            log.info("Request was cancelled: %s", url)
            # Don't set value to the request cache so we can query
            # later again

        except Exception as error:
            log.error("Error during request: %s", error)
            self._request_cache[url] = None

        else:
            self._request_cache[url] = og_data

        if og_data is None:
            log.warning("No opengraph data found for url: %s", url)

        self._update_preview(url, og_data)

    def _update_preview_list(self, urls: list[str]) -> None:
        self._remove_obsolete_previews(urls)
        self._add_previews(urls)

        for url in urls:
            if self._requests_in_progress.get(url):
                continue

            og_data = self._request_cache.get(url, MISSING)
            if isinstance(og_data, _MISSING):
                self._request_url_preview(url)
                continue

            self._update_preview(url, og_data)

        self.set_visible(bool(self._current_previews))

    def _add_previews(self, urls: list[str]) -> None:
        for url in urls:
            if url in self._current_previews:
                continue

            log.debug("Create preview widget for %s", url)
            preview = OpenGraphPreviewWidget(url)
            preview.connect("remove", self._on_remove_clicked, url)
            self.append(preview)
            self._current_previews[url] = preview

    def _remove_obsolete_previews(self, current_urls: list[str]) -> None:
        for url in list(self._current_previews.keys()):
            if url not in current_urls:
                preview = self._current_previews.pop(url)
                log.debug("Remove preview widget for %s", url)
                self.remove(preview)
                check_finalize(preview)

    def _update_preview(self, url: str, og_data: OpenGraphData | None) -> None:
        preview = self._current_previews.get(url)
        if preview is None:
            log.debug("No preview widget found")
            return

        if og_data is None:
            preview.set_error()
        else:
            preview.set_open_graph(og_data, minimal=True)

    def _on_remove_clicked(self, widget: OpenGraphPreviewWidget, url: str) -> None:
        self.remove(widget)
        self._current_previews.pop(url, None)
        self._dismissed_previews.add(url)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Literal

from gi.repository import Gtk

from gajim.gtk.chat_filter import ChatFilter
from gajim.gtk.menus import get_start_chat_button_menu
from gajim.gtk.util.misc import get_ui_string


@Gtk.Template(string=get_ui_string("chat_list_header.ui"))
class ChatListHeader(Gtk.Grid):
    __gtype_name__ = "ChatListHeader"

    _header_bar_label: Gtk.Label = Gtk.Template.Child()
    _chat_page_header: Gtk.Box = Gtk.Template.Child()
    _search_entry: Gtk.SearchEntry = Gtk.Template.Child()
    _chat_filter: ChatFilter = Gtk.Template.Child()
    _start_chat_menu_button: Gtk.MenuButton = Gtk.Template.Child()

    def __init__(self):
        Gtk.Grid.__init__(self)

        self._start_chat_menu_button.set_menu_model(get_start_chat_button_menu())

    def get_chat_filter(self) -> ChatFilter:
        return self._chat_filter

    def get_search_entry(self) -> Gtk.SearchEntry:
        return self._search_entry

    def set_label(self, text: str) -> None:
        self._header_bar_label.set_label(text)

    def set_header_mode(self, mode: Literal["chat", "activity"]) -> None:
        is_chat = mode == "chat"
        self._start_chat_menu_button.set_visible(is_chat)
        self._chat_filter.set_visible(is_chat)

        if is_chat:
            self._search_entry.set_text("")
            self._chat_filter.reset()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0047: In-Band Bytestreams

from __future__ import annotations

import time

import nbxmpp
from nbxmpp.errors import StanzaError
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import IqProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import types
from gajim.common.file_props import FileProp
from gajim.common.file_props import FilesProp
from gajim.common.helpers import to_user_string
from gajim.common.modules.base import BaseModule


class IBB(BaseModule):

    _nbxmpp_extends = 'IBB'
    _nbxmpp_methods = [
        'send_open',
        'send_close',
        'send_data',
        'send_reply',
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(name='iq',
                          callback=self._ibb_received,
                          ns=Namespace.IBB),
        ]

    def _ibb_received(self,
                      _con: types.NBXMPPClient,
                      stanza: Iq,
                      properties: IqProperties
                      ) -> None:
        if not properties.is_ibb:
            return

        if properties.ibb.type == 'data':
            self._log.info('Data received, sid: %s, seq: %s',
                           properties.ibb.sid, properties.ibb.seq)
            file_props = FilesProp.getFilePropByTransportSid(self._account,
                                                             properties.ibb.sid)
            if not file_props:
                self.send_reply(stanza, nbxmpp.ERR_ITEM_NOT_FOUND)
                raise NodeProcessed

            if file_props.connected:
                self._on_data_received(stanza, file_props, properties)
                self.send_reply(stanza)

        elif properties.ibb.type == 'open':
            self._log.info('Open received, sid: %s, blocksize: %s',
                           properties.ibb.sid, properties.ibb.block_size)

            file_props = FilesProp.getFilePropByTransportSid(self._account,
                                                             properties.ibb.sid)
            if not file_props:
                self.send_reply(stanza, nbxmpp.ERR_ITEM_NOT_FOUND)
                raise NodeProcessed

            file_props.block_size = properties.ibb.block_size
            file_props.direction = '<'
            file_props.seq = 0
            file_props.received_len = 0
            file_props.last_time = time.time()
            file_props.error = 0
            file_props.paused = False
            file_props.connected = True
            file_props.completed = False
            file_props.disconnect_cb = None
            file_props.continue_cb = None
            file_props.syn_id = stanza.getID()
            file_props.fp = open(file_props.file_name, 'wb')  # pylint: disable=consider-using-with  # noqa: E501
            self.send_reply(stanza)

        elif properties.ibb.type == 'close':
            self._log.info('Close received, sid: %s', properties.ibb.sid)
            file_props = FilesProp.getFilePropByTransportSid(self._account,
                                                             properties.ibb.sid)
            if not file_props:
                self.send_reply(stanza, nbxmpp.ERR_ITEM_NOT_FOUND)
                raise NodeProcessed

            self.send_reply(stanza)
            file_props.fp.close()
            file_props.completed = file_props.received_len >= file_props.size
            if not file_props.completed:
                file_props.error = -1
            app.socks5queue.complete_transfer_cb(self._account, file_props)

        raise NodeProcessed

    def _on_data_received(self,
                          stanza: Iq,
                          file_props: FileProp,
                          properties: IqProperties
                          ) -> None:
        ibb = properties.ibb
        if ibb.seq != file_props.seq:
            self.send_reply(stanza, nbxmpp.ERR_UNEXPECTED_REQUEST)
            self.send_close(file_props)
            raise NodeProcessed

        self._log.debug('Data received: sid: %s, %s+%s bytes',
                        ibb.sid, file_props.fp.tell(), len(ibb.data))

        file_props.seq += 1
        file_props.started = True
        file_props.fp.write(ibb.data)
        current_time = time.time()
        file_props.elapsed_time += current_time - file_props.last_time
        file_props.last_time = current_time
        file_props.received_len += len(ibb.data)
        app.socks5queue.progress_transfer_cb(self._account, file_props)
        if file_props.received_len >= file_props.size:
            file_props.completed = True

    def send_open(self, to: str, sid: str, fp: FileProp) -> FileProp:
        self._log.info('Send open to %s, sid: %s', to, sid)
        file_props = FilesProp.getFilePropBySid(sid)
        file_props.direction = '>'
        file_props.block_size = 4096
        file_props.fp = fp
        file_props.seq = -1
        file_props.error = 0
        file_props.paused = False
        file_props.received_len = 0
        file_props.last_time = time.time()
        file_props.connected = True
        file_props.completed = False
        file_props.disconnect_cb = None
        file_props.continue_cb = None
        self._nbxmpp('IBB').send_open(to,
                                      file_props.transport_sid,
                                      4096,
                                      callback=self._on_open_result,
                                      user_data=file_props)
        return file_props

    def _on_open_result(self, task: Task) -> None:
        try:
            task.finish()
        except StanzaError as error:
            app.socks5queue.error_cb('Error', to_user_string(error))
            self._log.warning(error)
            return

        file_props = task.get_user_data()
        self.send_data(file_props)

    def send_close(self, file_props: FileProp) -> None:
        file_props.connected = False
        file_props.fp.close()
        file_props.stopped = True
        to = file_props.receiver
        if file_props.direction == '<':
            to = file_props.sender

        self._log.info('Send close to %s, sid: %s',
                       to, file_props.transport_sid)
        self._nbxmpp('IBB').send_close(to, file_props.transport_sid,
                                       callback=self._on_close_result)

        if file_props.completed:
            app.socks5queue.complete_transfer_cb(self._account, file_props)
        else:
            if file_props.type_ == 's':
                peerjid = file_props.receiver
            else:
                peerjid = file_props.sender
            session = self._con.get_module('Jingle').get_jingle_session(
                peerjid, file_props.sid, 'file')
            # According to the xep, the initiator also cancels
            # the jingle session if there are no more files to send using IBB
            if session.weinitiate:
                session.cancel_session()

    def _on_close_result(self, task: Task) -> None:
        try:
            task.finish()
        except StanzaError as error:
            app.socks5queue.error_cb('Error', to_user_string(error))
            self._log.warning(error)
            return

    def send_data(self, file_props: FileProp) -> None:
        if file_props.completed:
            self.send_close(file_props)
            return

        chunk = file_props.fp.read(file_props.block_size)
        if chunk:
            file_props.seq += 1
            file_props.started = True
            if file_props.seq == 65536:
                file_props.seq = 0

            self._log.info('Send data to %s, sid: %s',
                           file_props.receiver, file_props.transport_sid)
            self._nbxmpp('IBB').send_data(file_props.receiver,
                                          file_props.transport_sid,
                                          file_props.seq,
                                          chunk,
                                          callback=self._on_data_result,
                                          user_data=file_props)
            current_time = time.time()
            file_props.elapsed_time += current_time - file_props.last_time
            file_props.last_time = current_time
            file_props.received_len += len(chunk)
            if file_props.size == file_props.received_len:
                file_props.completed = True
            app.socks5queue.progress_transfer_cb(self._account, file_props)

    def _on_data_result(self, task: Task) -> None:
        try:
            task.finish()
        except StanzaError as error:
            app.socks5queue.error_cb('Error', to_user_string(error))
            self._log.warning(error)
            return

        file_props = task.get_user_data()
        self.send_data(file_props)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import multiprocessing
import os
import platform
import signal
import sqlite3
import sys
import warnings
from ctypes import byref
from ctypes import CDLL
from ctypes import create_string_buffer
from ctypes.util import find_library
from types import FrameType

from packaging.version import Version as V

_MIN_NBXMPP_VER = "7.1.0"
_MIN_GTK_VER = "4.17.5"
_MIN_ADW_VER = "1.7.0"
_MIN_CAIRO_VER = "1.16.0"
_MIN_PYGOBJECT_VER = "3.53.0"
_MIN_GLIB_VER = "2.80.0"
_MIN_PANGO_VER = "1.50.0"
_MIN_SQLITE_VER = "3.35.0"

error_message = "Gajim needs {0} to run. Quitting… (Error: {1})"


def check_version(dep_name: str, current_ver: str, min_ver: str) -> None:
    if V(current_ver) < V(min_ver):
        sys.exit(
            f"Gajim needs {dep_name} >= {min_ver} "
            f"(found {current_ver}) to run. Quitting..."
        )


def gi_require_versions() -> None:
    import gi

    gi.require_versions(
        {
            "Adw": "1",
            "Gdk": "4.0",
            "GdkPixbuf": "2.0",
            "Gio": "2.0",
            "GLib": "2.0",
            "GObject": "2.0",
            "Graphene": "1.0",
            "Gsk": "4.0",
            "Gtk": "4.0",
            "GtkSource": "5",
            "Pango": "1.0",
            "PangoCairo": "1.0",
        }
    )


def _check_required_deps() -> None:
    try:
        import gi
    except ImportError as error:
        sys.exit(error_message.format("PyGObject", error))

    check_version("pygobject", gi.__version__, _MIN_PYGOBJECT_VER)

    gi.disable_legacy_autoinit()

    try:
        import nbxmpp
    except ImportError as error:
        sys.exit(error_message.format("python-nbxmpp", error))

    try:
        gi_require_versions()
    except ValueError as error:
        sys.exit(f"Missing dependency: {error}")

    try:
        import cairo
    except ImportError as error:
        sys.exit(error_message.format("pycairo", error))

    from gi.repository import Gtk

    gtk_ver = (
        f"{Gtk.get_major_version()}.{Gtk.get_minor_version()}.{Gtk.get_micro_version()}"
    )

    from gi.repository import Adw

    adw_ver = (
        f"{Adw.get_major_version()}.{Adw.get_minor_version()}.{Adw.get_micro_version()}"
    )

    from gi.repository import GLib

    glib_ver = f"{GLib.MAJOR_VERSION}.{GLib.MINOR_VERSION}.{GLib.MICRO_VERSION}"

    from gi.repository import Pango

    check_version("python-nbxmpp", nbxmpp.__version__, _MIN_NBXMPP_VER)
    check_version("libcairo", cairo.cairo_version_string(), _MIN_CAIRO_VER)
    check_version("pycairo", cairo.version, _MIN_CAIRO_VER)
    check_version("gtk4", gtk_ver, _MIN_GTK_VER)
    check_version("adw", adw_ver, _MIN_ADW_VER)
    check_version("glib", glib_ver, _MIN_GLIB_VER)
    check_version("pango", Pango.version_string(), _MIN_PANGO_VER)
    check_version("sqlite", sqlite3.sqlite_version, _MIN_SQLITE_VER)


def _init_translations() -> None:
    import gajim.common.i18n

    gajim.common.i18n.init()


def _init_gui(gui: str) -> None:
    if gui == "GTK":
        _init_gtk()


def _set_env_vars() -> None:
    # Disable legacy ciphers in cryptography
    os.environ["CRYPTOGRAPHY_OPENSSL_NO_LEGACY"] = "1"

    if sys.platform != "win32":
        return

    # Disable accesskit due to segfaults when calling app.window.set_visible(True)
    # https://github.com/msys2/MINGW-packages/issues/24812
    os.environ["GTK_A11Y"] = "none"

    if "GTK_CSD" in os.environ:
        # Respect user settings
        return

    os.environ["GTK_CSD"] = "0"


def _init_gtk() -> None:
    from gajim.gtk import exception

    exception.init()


def _run_app() -> None:
    from gajim.gtk.application import GajimApplication

    application = GajimApplication()

    def sigint_cb(num: int, stack: FrameType | None) -> None:
        print(" SIGINT/SIGTERM received")
        application.start_shutdown()

    # ^C exits the application normally
    signal.signal(signal.SIGINT, sigint_cb)
    signal.signal(signal.SIGTERM, sigint_cb)
    if sys.platform != "win32":
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)

    # https://gitlab.gnome.org/GNOME/pygobject/-/issues/12
    warnings.filterwarnings("ignore", ".*g_value_get_int.*G_VALUE_HOLDS_INT", Warning)

    application.run(sys.argv)


def set_proc_title(title: str) -> None:
    sysname = platform.system()
    if sysname in ("Linux", "FreeBSD", "OpenBSD", "NetBSD"):
        libc = CDLL(find_library("c"))

        # The constant defined in <linux/prctl.h> which is used to set the name
        # of the process.
        PR_SET_NAME = 15

        if sysname == "Linux":
            proc_name = title.encode("utf-8")
            buff = create_string_buffer(len(proc_name) + 1)
            buff.value = proc_name
            libc.prctl(PR_SET_NAME, byref(buff), 0, 0, 0)
        elif sysname in ("FreeBSD", "OpenBSD", "NetBSD"):
            libc.setproctitle(title)


def run() -> None:
    multiprocessing.freeze_support()
    if sys.platform != "win32":
        if os.geteuid() == 0:
            sys.exit("You must not launch gajim as root, it is insecure.")

    _check_required_deps()
    set_proc_title("gajim")
    _set_env_vars()
    _init_translations()
    _init_gui("GTK")
    _run_app()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import datetime as dt
import logging

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango
from nbxmpp import JID

from gajim.common import app
from gajim.common.const import AvatarSize
from gajim.common.i18n import _
from gajim.common.storage.archive import models as mod
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageType
from gajim.common.structs import ReplyData
from gajim.common.types import ChatContactT
from gajim.common.util.preview import get_preview_data
from gajim.common.util.text import quote_text

from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_avatar_for_message
from gajim.gtk.util.misc import get_contact_name_for_message

log = logging.getLogger("gajim.gtk.referenced_message_widget")


class ReferencedMessageWidget(Gtk.Box, SignalManager):
    def __init__(
        self,
        contact: ChatContactT,
        original_message: mod.Message,
        show_reply_icon: bool = True,
    ) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self._contact = contact
        self._original_message = original_message
        self._message = original_message.get_last_correction() or original_message
        self._show_reply_icon = show_reply_icon

        self.set_cursor(Gdk.Cursor.new_from_name("pointer"))

        gesture_primary_click = Gtk.GestureClick(button=Gdk.BUTTON_PRIMARY)
        self._connect(gesture_primary_click, "pressed", self._on_clicked)
        self.add_controller(gesture_primary_click)

        self._add_content(self._message)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Box.do_unroot(self)

    def _add_content(self, message: mod.Message) -> None:
        main_box = Gtk.Box(
            spacing=12, hexpand=True, tooltip_text=_("Scroll to this message")
        )
        main_box.add_css_class("referenced-message")
        self.append(main_box)

        quote_bar = Gtk.Box(width_request=4)
        quote_bar.set_name("quote-bar")
        main_box.append(quote_bar)

        content_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL, halign=Gtk.Align.START
        )
        main_box.append(content_box)

        avatar_surface = get_avatar_for_message(
            message, self._contact, self.get_scale_factor(), AvatarSize.MESSAGE_REPLY
        )
        avatar_image = Gtk.Image.new_from_paintable(avatar_surface)
        avatar_image.set_pixel_size(AvatarSize.MESSAGE_REPLY)

        name = get_contact_name_for_message(message, self._contact)
        name_label = Gtk.Label(label=name)
        name_label.add_css_class("dimmed")
        name_label.add_css_class("bold")

        reply_icon = Gtk.Image.new_from_icon_name("lucide-reply-symbolic")

        timestamp = message.timestamp.astimezone()
        format_string = app.settings.get("time_format")
        if timestamp.date() < dt.datetime.today().date():
            format_string = app.settings.get("date_time_format")

        timestamp_label = Gtk.Label(
            label=timestamp.strftime(format_string),
            margin_start=6,
        )
        timestamp_label.add_css_class("dimmed")

        meta_box = Gtk.Box(spacing=6, valign=Gtk.Align.CENTER)
        meta_box.append(reply_icon)
        meta_box.append(avatar_image)
        meta_box.append(name_label)
        meta_box.append(timestamp_label)
        content_box.append(meta_box)

        message_box = Gtk.Box(spacing=12)
        label_text = ""
        if message.text is not None:
            preview = get_preview_data(message.text, message.oob)
            if preview is not None:
                label_text = preview.text
                image = Gtk.Image.new_from_gicon(preview.icon)
                message_box.append(image)

            else:
                label_text = message.text
                lines = message.text.split("\n")
                if len(lines) > 3:
                    label_text = "\n".join(lines[:3])
                    label_text = f"{label_text} …"

        message_label = Gtk.Label(
            label=label_text,
            halign=Gtk.Align.START,
            max_width_chars=100,
            ellipsize=Pango.EllipsizeMode.END,
        )
        message_label.add_css_class("dimmed")
        message_label.add_css_class("document")

        message_box.append(message_label)
        content_box.append(message_box)

    def _on_clicked(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        _x: float,
        _y: float,
    ) -> None:
        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
        app.window.activate_action(
            "win.jump-to-message",
            GLib.Variant(
                "au",
                [
                    self._message.pk,
                    self._message.timestamp.timestamp(),
                ],
            ),
        )

    def get_message_reply(self) -> ReplyData | None:
        # We only show the reply menu if there is text
        assert self._message.text

        jid = self._message.remote.jid
        reply_to_id = self._original_message.id

        if self._message.type == MessageType.GROUPCHAT:
            jid = self._message.remote.jid
            jid = jid.new_with(resource=self._message.resource)
            reply_to_id = self._original_message.stanza_id
        elif self._message.type == MessageType.CHAT:
            if self._original_message.direction == ChatDirection.OUTGOING:
                jid = JID.from_string(app.get_jid_from_account(self._contact.account))

        if reply_to_id is None:
            return None

        thread_id = None
        if self._original_message.thread is not None:
            thread_id = self._original_message.thread.id

        quoted_text = quote_text(self._message.text)
        return ReplyData(
            pk=self._original_message.pk,
            to=jid,
            id=reply_to_id,
            thread_id=thread_id,
            fallback_start=0,
            fallback_end=len(quoted_text),
            fallback_text=quoted_text,
        )


class ReferencedMessageNotFoundWidget(Gtk.Box):
    def __init__(self) -> None:
        Gtk.Box.__init__(self)

        main_box = Gtk.Box(spacing=12, hexpand=True)
        main_box.add_css_class("referenced-message")
        self.append(main_box)

        quote_bar = Gtk.Box(width_request=4)
        quote_bar.set_name("quote-bar")
        main_box.append(quote_bar)

        content_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL, halign=Gtk.Align.START
        )
        main_box.append(content_box)

        message_box = Gtk.Box(spacing=12)
        message_label = Gtk.Label(
            label=_("The referenced message is not available."),
            halign=Gtk.Align.START,
            max_width_chars=100,
            ellipsize=Pango.EllipsizeMode.END,
        )
        message_label.add_css_class("dimmed")

        message_box.append(message_label)
        content_box.append(message_box)


class ReplyBox(Gtk.Box, SignalManager):
    def __init__(self) -> None:
        Gtk.Box.__init__(self, spacing=6, visible=False)
        SignalManager.__init__(self)

        reply_image = Gtk.Image.new_from_icon_name("lucide-reply-symbolic")
        reply_image.set_size_request(AvatarSize.CHAT, -1)
        reply_image.set_pixel_size(24)
        reply_image.add_css_class("dimmed")
        self.append(reply_image)

        self._close_button = Gtk.Button.new_from_icon_name("lucide-x-symbolic")
        self._close_button.set_valign(Gtk.Align.CENTER)
        self._close_button.set_tooltip_text(_("Cancel"))
        self._connect(self._close_button, "clicked", self.disable_reply_mode)
        self.append(self._close_button)

        self._ref_widget = None

    def do_unroot(self) -> None:
        self.disable_reply_mode()
        self._disconnect_all()
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    def enable_reply_mode(
        self, contact: ChatContactT, original_message: mod.Message
    ) -> None:
        if self._ref_widget is not None:
            self.disable_reply_mode()

        self._ref_widget = ReferencedMessageWidget(
            contact, original_message, show_reply_icon=False
        )
        self.append(self._ref_widget)
        self.reorder_child_after(self._close_button, self._ref_widget)
        self.set_visible(True)

    def disable_reply_mode(self, *args: Any) -> None:
        if self._ref_widget is not None:
            self.remove(self._ref_widget)
            self._ref_widget = None

        self.set_visible(False)

    @property
    def is_in_reply_mode(self) -> bool:
        return self._ref_widget is not None

    def get_message_reply(self) -> ReplyData | None:
        if self._ref_widget is None:
            return None

        return self._ref_widget.get_message_reply()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.util.datetime import FIRST_LOCAL_DATETIME

from gajim.gtk.util.classes import SignalManager


class BaseRow(Gtk.ListBoxRow, SignalManager):
    __gsignals__ = {
        "remove": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, account: str, widget: str | None = None) -> None:
        Gtk.ListBoxRow.__init__(self, selectable=False)
        SignalManager.__init__(self)

        self._account = account
        self._client = app.get_client(account)

        self.type: str = ""
        self.timestamp = FIRST_LOCAL_DATETIME
        self.kind: str = ""
        self.direction = ChatDirection.INCOMING
        self.name: str = ""
        self.pk: int | None = None
        self.stanza_id: str | None = None
        self.text: str = ""
        self._merged: bool = False

        self.add_css_class("conversation-row")

        self.grid = Gtk.Grid(row_spacing=3, column_spacing=12)
        self.set_child(self.grid)

        if widget == "label":
            self.label = Gtk.Label(
                selectable=True, wrap=True, xalign=0, wrap_mode=Pango.WrapMode.WORD_CHAR
            )

    def enable_selection_mode(self) -> None:
        return

    def disable_selection_mode(self) -> None:
        return

    @property
    def is_merged(self) -> bool:
        return self._merged

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.ListBoxRow.do_unroot(self)



# Copyright (C) 2005-2006 Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2005-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2006 Jean-Marie Traissard <jim AT lapin.org>
# Copyright (C) 2007 Brendan Taylor <whateley AT gmail.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only


class GajimGeneralException(Exception):
    '''
    This exception is our general exception
    '''

    def __init__(self, text: str = '') -> None:
        Exception.__init__(self)
        self.text = text

    def __str__(self) -> str:
        return self.text


class PluginsystemError(Exception):
    '''
    Error in the pluginsystem
    '''

    def __init__(self, text: str = '') -> None:
        Exception.__init__(self)
        self.text = text

    def __str__(self) -> str:
        return self.text


class FileError(Exception):
    pass



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import cast

import logging

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.dsa import DSAPublicKey
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.utils import int_to_bytes
from cryptography.x509 import DNSName
from cryptography.x509 import ExtensionNotFound
from cryptography.x509.oid import ExtensionOID
from gi.repository import Adw
from gi.repository import Gio
from gi.repository import Gtk

from gajim.common import app
from gajim.common.helpers import get_x509_cert_from_gio_cert
from gajim.common.i18n import _
from gajim.common.util.text import format_bytes_as_hex
from gajim.common.util.version import package_version

from gajim.gtk.preference.widgets import CopyButton
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string

log = logging.getLogger("gajim.gtk.preferences.certificate")


@Gtk.Template.from_string(string=get_ui_string("preference/certificate.ui"))
class CertificatePage(Adw.PreferencesPage, SignalManager):
    __gtype_name__ = "CertificatePage"

    _clipboard_button: CopyButton = Gtk.Template.Child()
    _account_label_row: Adw.ActionRow = Gtk.Template.Child()
    _account_address_row: Adw.ActionRow = Gtk.Template.Child()
    _issued_to_common_name_row: Adw.ActionRow = Gtk.Template.Child()
    _issued_to_organization_row: Adw.ActionRow = Gtk.Template.Child()
    _issued_to_subject_alt_names_row: Adw.ActionRow = Gtk.Template.Child()
    _issued_to_serial_number_row: Adw.ActionRow = Gtk.Template.Child()
    _issued_by_common_name_row: Adw.ActionRow = Gtk.Template.Child()
    _issued_by_organization_row: Adw.ActionRow = Gtk.Template.Child()
    _issue_date_row: Adw.ActionRow = Gtk.Template.Child()
    _expiry_date_row: Adw.ActionRow = Gtk.Template.Child()
    _fingerprint_sha1_row: Adw.ActionRow = Gtk.Template.Child()
    _fingerprint_sha256_row: Adw.ActionRow = Gtk.Template.Child()
    _algorithm_row: Adw.ActionRow = Gtk.Template.Child()
    _key_size_row: Adw.ActionRow = Gtk.Template.Child()

    def __init__(self, account: str | None, certificate: Gio.TlsCertificate) -> None:
        Adw.PreferencesPage.__init__(self)
        SignalManager.__init__(self)

        self._connect(self._clipboard_button, "clicked", self._on_copy_button_clicked)

        self._account_label = ""
        self._account_address = ""
        if account is not None:
            self._account_label = app.settings.get_account_setting(
                account, "account_label"
            )
            self._account_address = app.settings.get_account_setting(account, "address")

        cert = get_x509_cert_from_gio_cert(certificate)

        self._it_common_name = "-"
        self._it_organization = "-"
        for attribute in cert.subject:
            # See https://datatracker.ietf.org/doc/html/rfc4514.html
            dotted_string = attribute.oid.dotted_string
            if dotted_string == "2.5.4.3":
                self._it_common_name = str(attribute.value)
            if dotted_string == "2.5.4.10":
                self._it_organization = str(attribute.value)

        # Get the subjectAltName extension from the certificate
        try:
            subject_ext = cert.extensions.get_extension_for_oid(
                ExtensionOID.SUBJECT_ALTERNATIVE_NAME
            )
            # Get the DNSName entries from the SAN extension
            alt_names = cast(
                list[str],
                subject_ext.value.get_values_for_type(DNSName),  # pyright: ignore
            )
            self._it_subject_alt_names = "\n".join(alt_names)
        except ExtensionNotFound as err:
            log.info("Certificate does not have extension: %s", err)
            self._it_subject_alt_names = "-"

        try:
            serial_bytes = format_bytes_as_hex(int_to_bytes(cert.serial_number), 2)
        except Exception:
            log.exception("Error while trying to get serial number of certificate")
            serial_bytes = ""
        self._it_serial_number = serial_bytes

        self._ib_common_name = "-"
        self._ib_organization = "-"
        for attribute in cert.issuer:
            # See https://datatracker.ietf.org/doc/html/rfc4514.html
            dotted_string = attribute.oid.dotted_string
            if dotted_string == "2.5.4.3":
                self._ib_common_name = str(attribute.value)
            if dotted_string == "2.5.4.10":
                self._ib_organization = str(attribute.value)

        if package_version("cryptography>=42.0.0"):
            self._issued = str(cert.not_valid_before_utc.strftime("%c %Z"))
            self._expires = str(cert.not_valid_after_utc.strftime("%c %Z"))
        else:
            self._issued = str(cert.not_valid_before.strftime("%c %Z"))
            self._expires = str(cert.not_valid_after.strftime("%c %Z"))

        sha1_bytes = cert.fingerprint(hashes.SHA1())
        self._sha1 = format_bytes_as_hex(sha1_bytes, 2)

        sha256_bytes = cert.fingerprint(hashes.SHA256())
        self._sha256 = format_bytes_as_hex(sha256_bytes, 4)

        public_key = cert.public_key()
        self._pk_algorithm = "-"
        if isinstance(public_key, RSAPublicKey):
            self._pk_algorithm = "RSA"
        elif isinstance(public_key, DSAPublicKey):
            self._pk_algorithm = "DSA"
        elif isinstance(public_key, EllipticCurvePublicKey):
            self._pk_algorithm = "Elliptic Curve"
        elif isinstance(public_key, Ed25519PublicKey):
            self._pk_algorithm = "ED25519"
        elif isinstance(public_key, Ed448PublicKey):
            self._pk_algorithm = "ED448"

        self._pk_size = _("Unknown")
        if isinstance(public_key, RSAPublicKey | DSAPublicKey | EllipticCurvePublicKey):
            self._pk_size = f"{public_key.key_size} Bit"

        self._account_label_row.set_subtitle(self._account_label)
        self._account_address_row.set_subtitle(self._account_address)

        self._issued_to_common_name_row.set_subtitle(self._it_common_name)
        self._issued_to_organization_row.set_subtitle(self._it_organization)
        self._issued_to_subject_alt_names_row.set_subtitle(self._it_subject_alt_names)
        self._issued_to_serial_number_row.set_subtitle(self._it_serial_number)

        self._issued_by_common_name_row.set_subtitle(self._ib_common_name)
        self._issued_by_organization_row.set_subtitle(self._ib_organization)

        self._issue_date_row.set_subtitle(self._issued)
        self._expiry_date_row.set_subtitle(self._expires)

        self._fingerprint_sha1_row.set_subtitle(self._sha1)
        self._fingerprint_sha256_row.set_subtitle(self._sha256)

        self._algorithm_row.set_subtitle(self._pk_algorithm)
        self._key_size_row.set_subtitle(self._pk_size)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Adw.PreferencesPage.do_unroot(self)
        app.check_finalize(self)

    def _on_copy_button_clicked(self, _widget: Gtk.Button) -> None:
        clipboard_text = (
            _("Certificate")
            + "\n"
            + _("Account name: ")
            + self._account_label
            + "\n"
            + _("Account address: ")
            + self._account_address
            + "\n"
            + _("Issued to\n")
            + _("Common Name (CN): ")
            + self._it_common_name
            + "\n"
            + _("Organization (O): ")
            + self._it_organization
            + "\n"
            + _("Subject Alt Names: ")
            + self._it_subject_alt_names
            + "\n"
            + _("Serial Number: ")
            + self._it_serial_number
            + "\n\n"
            + _("Issued by\n")
            + _("Common Name (CN): ")
            + self._ib_common_name
            + "\n"
            + _("Organization (O): ")
            + self._ib_organization
            + "\n"
            + _("Validity\n")
            + _("Issued on: ")
            + self._issued
            + "\n"
            + _("Expires on: ")
            + self._expires
            + "\n\n"
            + _("SHA-1:")
            + "\n"
            + self._sha1
            + "\n"
            + _("SHA-256:")
            + "\n"
            + self._sha256
            + "\n\n"
            + _("Public Key: ")
            + self._pk_algorithm
            + " "
            + self._pk_size
        )

        app.window.get_clipboard().set(clipboard_text)



# Copyright (C) 2003-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2005-2006 Dimitur Kirov <dkirov AT gmail.com>
#                         Travis Shirk <travis AT pobox.com>
#                         Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2006 Junglecow J <junglecow AT gmail.com>
#                    Stefan Bethge <stefan AT lanpartei.de>
# Copyright (C) 2006-2008 Jean-Marie Traissard <jim AT lapin.org>
# Copyright (C) 2007-2008 Brendan Taylor <whateley AT gmail.com>
#                         Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2008 Jonathan Schleifer <js-gajim AT webkeks.org>
# Copyright (C) 2018 Philipp Hörist <philipp @ hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing
from typing import Any
from typing import cast
from typing import Union

import gc
import logging
import multiprocessing
import os
import pprint
import sys
import weakref
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from nbxmpp.protocol import JID
from nbxmpp.task import Task

import gajim
from gajim.common import configpaths
from gajim.common import ged as ged_module
from gajim.common import types
from gajim.common.const import Display
from gajim.common.i18n import get_default_lang
from gajim.common.multiprocess import init_process

if typing.TYPE_CHECKING:
    from gajim.common.call_manager import CallManager
    from gajim.common.cert_store import CertificateStore
    from gajim.common.commands import ChatCommands  # noqa: F401
    from gajim.common.file_transfer_manager import FileTransferManager
    from gajim.common.storage.archive.storage import MessageArchiveStorage
    from gajim.common.storage.cache import CacheStorage
    from gajim.common.storage.draft import DraftStorage
    from gajim.common.storage.events.storage import EventStorage
    from gajim.common.task_manager import PulseManager
    from gajim.common.task_manager import TaskManager

    from gajim.gtk.application import GajimApplication  # noqa: F401
    from gajim.gtk.audio_player import AudioPlayer
    from gajim.gtk.main import MainWindow  # noqa: F401


settings = cast(types.SettingsT, None)
version = gajim.__version__
connections: dict[str, types.Client] = {}
avatar_cache: dict[str, dict[str, Any]] = {}
bob_cache: dict[str, bytes] = {}
services_cache: dict[str, Any] = {}
app = None  # type: GajimApplication
window = None  # type: MainWindow
commands = None  # type: ChatCommands
logging_records: list[str] = []

ged = ged_module.GlobalEventsDispatcher()  # Global Events Dispatcher
plugin_manager = cast(types.PluginManagerT, None)  # Plugins Manager
plugin_repository = cast(types.PluginRepositoryT, None)


class Storage:
    def __init__(self):
        self.cache: CacheStorage = None
        self.archive: MessageArchiveStorage = None
        self.events: EventStorage = None
        self.drafts: DraftStorage = None


storage = Storage()

css_config = cast(types.CSSConfigT, None)

transport_type: dict[str, str] = {}

# list of contacts that has just signed out
to_be_removed: dict[str, list[str]] = {}

notification = None

# list of our nick names in each account
nicks: dict[str, str] = {}

cert_store = cast('CertificateStore', None)

call_manager = cast('CallManager', None)
audio_player: Union['AudioPlayer', None] = None  # noqa: UP007
ftm = cast('FileTransferManager', None)

task_manager = cast('TaskManager', None)
pulse_manager = cast('PulseManager', None)

gupnp_igd = None

gsound_ctx = None

process_pool = cast(ProcessPoolExecutor, None)

_dependencies = {
    'FARSTREAM': False,
    'GST': False,
    'AV': False,
    'GEOCLUE': False,
    'UPNP': False,
    'GSOUND': False,
    'SPELLING': False,
    'IDLE': False,
    'SENTRY_SDK': False,
}

_tasks: dict[int, list[Task]] = defaultdict(list)


def init_process_pool() -> None:
    global process_pool

    mp_context = multiprocessing.get_context("spawn")
    process_pool = ProcessPoolExecutor(
        max_workers=8,
        mp_context=mp_context,
        # https://github.com/python/cpython/issues/115634
        # Start a new process per task to ensure memory is released,
        # when generating video thumbnails using Gstreamer
        max_tasks_per_child=1,
        initializer=init_process,
    )


def print_version() -> None:
    log('gajim').info('Gajim Version: %s', gajim.__version__)


def get_client(account: str) -> types.Client:
    return connections[account]


def get_clients() -> list[types.Client]:
    return list(connections.values())


def is_installed(dependency: str) -> bool:
    return _dependencies[dependency]


def is_flatpak() -> bool:
    return gajim.IS_FLATPAK


def is_flatpak_nightly() -> bool:
    return gajim.IS_FLATPAK_NIGHTLY


def is_portable() -> bool:
    return gajim.IS_PORTABLE


def is_ms_store() -> bool:
    return gajim.IS_MS_STORE


def get_default_app_id() -> str:
    app_id = 'org.gajim.Gajim'
    if is_flatpak_nightly():
        return f'{app_id}.Devel'
    return app_id


def is_display(display: Display) -> bool:
    # XWayland reports as Display X11, so try with env var
    is_wayland = os.environ.get('XDG_SESSION_TYPE') == 'wayland'
    if is_wayland and display == Display.WAYLAND:
        return True

    default = Gdk.Display.get_default()
    if default is None:
        log('gajim').warning('Could not determine window manager')
        return False
    return default.__class__.__name__ == display.value


def disable_dependency(dependency: str) -> None:
    _dependencies[dependency] = False


def detect_dependencies() -> None:
    import gi

    try:
        gi.require_version('Gst', '1.0')
        gi.require_version('GstPbutils', '1.0')
        from gi.repository import Gst
        from gi.repository import GstPbutils  # noqa: F401
        success, _argv = Gst.init_check(None)
        _dependencies['GST'] = success
    except Exception:
        pass

    try:
        gi.require_version('Farstream', '0.2')
        # from gi.repository import Farstream
        # _dependencies['FARSTREAM'] = True
    except Exception as error:
        log('gajim').warning('AV dependency test failed: %s', error)

    # try:
    #     if _dependencies['GST'] and _dependencies['FARSTREAM']:
    #         # This creates a GTK critical
    #         # gst_base_auto_convert_register_filter: assertion
    #         # 'g_strcmp0 (name, ((GstAutoConvertFilterInfo *) tmp->data)->name)'
    #         # failed
    #         conference = Gst.ElementFactory.make('fsrtpconference', None)
    #         conference.new_session(Farstream.MediaType.AUDIO)
    #         from gajim.gtk.gstreamer import create_video_elements
    #         video_elements = create_video_elements()
    #         if video_elements is not None:
    #             _dependencies['AV'] = True
    # except Exception as error:
    #     log('gajim').warning('AV dependency test failed: %s', error)

    # GEOCLUE
    try:
        gi.require_version('Geoclue', '2.0')
        from gi.repository import Geoclue  # noqa: F401
        _dependencies['GEOCLUE'] = True
    except (ImportError, ValueError):
        pass

    # UPNP
    # GUPnPidg uses libsoup3 which is incompatible with Gajim.
    # Don’t load until Gajim is ported to libsoup3
    #
    # try:
    #     gi.require_version('GUPnPIgd', '1.0')
    #     from gi.repository import GUPnPIgd
    #     global gupnp_igd  # pylint: disable=global-statement
    #     gupnp_igd = GUPnPIgd.SimpleIgd()
    #     _dependencies['UPNP'] = True
    # except ValueError:
    #     pass

    # IDLE
    try:
        from gajim.common import idle
        if idle.Monitor.is_available():
            _dependencies['IDLE'] = True
    except Exception:
        pass

    # GSOUND
    try:
        gi.require_version('GSound', '1.0')
        from gi.repository import GSound
        global gsound_ctx  # pylint: disable=global-statement
        gsound_ctx = GSound.Context()
        try:
            gsound_ctx.init()
            _dependencies['GSOUND'] = True
        except GLib.Error as error:
            log('gajim').warning('GSound init failed: %s', error)
    except (ImportError, ValueError):
        pass

    # LIBSPELLING
    try:
        gi.require_version('Spelling', '1')
        from gi.repository import Spelling
        provider = Spelling.Provider.get_default()
        langs = provider.list_languages()
        for lang in langs:
            log('gajim').info('%s (%s) dict available',
                              lang.get_name(), lang.get_code())
        if langs:
            _dependencies['SPELLING'] = True
    except (ImportError, ValueError):
        pass

    # SENTRY SDK
    try:
        import sentry_sdk  # noqa: F401
        _dependencies['SENTRY_SDK'] = True
    except Exception:
        # Sentry has a lot of side effects on import
        # make sure this optional dependency does not prevent
        # Gajim from starting
        pass

    # Print results
    for dep, val in _dependencies.items():
        log('gajim').info('%-13s %s', dep, val)

    log('gajim').info('Used language: %s', get_default_lang())


def detect_desktop_env() -> str | None:
    if sys.platform in ('win32', 'darwin'):
        return sys.platform

    desktop = os.environ.get('XDG_CURRENT_DESKTOP')
    if desktop is None:
        return None

    if 'gnome' in desktop.lower():
        return 'gnome'
    return desktop


desktop_env = detect_desktop_env()


def get_server_from_jid(jid: str) -> str:
    pos = jid.find('@') + 1  # after @
    return jid[pos:]


def get_room_and_nick_from_fjid(jid: str) -> list[str]:
    # fake jid is the jid for a contact in a room
    # gaim@conference.jabber.no/nick/nick-continued
    # return ('gaim@conference.jabber.no', 'nick/nick-continued')
    list_ = jid.split('/', 1)
    if len(list_) == 1:  # No nick
        list_.append('')
    return list_


def get_jid_without_resource(jid: str) -> str:
    return jid.split('/', maxsplit=1)[0]


def get_number_of_connected_accounts(
        accounts_list: list[str] | None = None) -> int:
    '''
    Returns the number of connected accounts. You can optionally pass an
    accounts_list and if you do those will be checked, else all will be checked
    '''
    connected_accounts = 0
    if accounts_list is None:
        accounts = connections.keys()
    else:
        accounts = accounts_list
    for account in accounts:
        if account_is_connected(account):
            connected_accounts = connected_accounts + 1
    return connected_accounts


def get_available_clients() -> list[types.Client]:
    clients: list[types.Client] = []
    for client in connections.values():
        if client.state.is_available:
            clients.append(client)
    return clients


def get_connected_accounts(exclude_local: bool = False) -> list[str]:
    '''
    Returns a list of CONNECTED accounts
    '''
    account_list: list[str] = []
    for account in connections:
        if account == 'Local' and exclude_local:
            continue
        if account_is_connected(account):
            account_list.append(account)
    return account_list


def get_accounts_sorted() -> list[str]:
    '''
    Get all accounts alphabetically sorted with Local first
    '''
    account_list = settings.get_accounts()
    account_list.sort(key=str.lower)
    return account_list


def get_enabled_accounts_with_labels(
    connected_only: bool = False
) -> list[list[str]]:
    '''
    Returns a list with [account, account_label] entries.
    Order by account_label
    '''
    accounts: list[list[str]] = []
    for acc in connections:
        if connected_only and not account_is_connected(acc):
            continue

        accounts.append([acc, get_account_label(acc)])

    accounts.sort(key=lambda xs: str.lower(xs[1]))
    return accounts


def get_account_label(account: str) -> str:
    return settings.get_account_setting(account, 'account_label') or account


def account_is_connected(account: str) -> bool:
    if account not in connections:
        return False
    return (connections[account].state.is_connected or
            connections[account].state.is_available)


def account_is_available(account: str) -> bool:
    if account not in connections:
        return False
    return connections[account].state.is_available


def get_transport_name_from_jid(
        jid: str,
        use_config_setting: bool = True) -> str | None:

    '''
    Returns 'gg', 'irc' etc

    If JID is not from transport returns None.
    '''
    # TODO: Rewrite/remove

    # FIXME: jid can be None! one TB I saw had this problem:
    # in the code block # it is a groupchat presence in handle_event_notify
    # jid was None. Yann why?
    if not jid:
        return

    host = get_server_from_jid(jid)
    if host in transport_type:
        return transport_type[host]

    # host is now f.e. icq.foo.org or just icq (sometimes on hacky transports)
    host_splitted = host.split('.')
    if host_splitted:
        # now we support both 'icq.' and 'icq' but not icqsucks.org
        host = host_splitted[0]

    if host in ('irc', 'icq', 'sms', 'weather', 'mrim', 'facebook'):
        return host
    if host == 'gg':
        return 'gadu-gadu'
    if host == 'jit':
        return 'icq'
    if host == 'facebook':
        return 'facebook'
    return None


def jid_is_transport(jid: str) -> bool:
    # if not '@' or '@' starts the jid then it is transport
    return jid.find('@') <= 0


def get_jid_from_account(account_name: str) -> str:
    '''
    Return the jid we use in the given account
    '''
    return settings.get_account_setting(account_name, 'address')


def get_default_nick(account_name: str) -> str:
    address = settings.get_account_setting(account_name, 'address')
    jid = JID.from_string(address)
    assert jid.localpart is not None
    return jid.localpart


def get_hostname_from_account(
    account_name: str,
    prefer_custom: bool = False
) -> str:
    if prefer_custom:
        if settings.get_account_setting(account_name, 'use_custom_host'):
            return settings.get_account_setting(account_name, 'custom_host')

    address = settings.get_account_setting(account_name, 'address')
    jid = JID.from_string(address)
    assert jid.domain is not None
    return jid.domain


def log(domain: str) -> logging.Logger:
    if domain != 'gajim':
        domain = 'gajim.%s' % domain
    return logging.getLogger(domain)


def load_css_config() -> None:
    global css_config   # pylint: disable=global-statement
    from gajim.gtk.css_config import CSSConfig
    css_config = CSSConfig()


def set_debug_mode(enable: bool) -> None:
    debug_folder = configpaths.get('DEBUG')
    debug_enabled = debug_folder / 'debug-enabled'
    if enable:
        debug_enabled.touch()
    else:
        if debug_enabled.exists():
            debug_enabled.unlink()


def get_debug_mode() -> bool:
    debug_folder = configpaths.get('DEBUG')
    debug_enabled = debug_folder / 'debug-enabled'
    return debug_enabled.exists()


def get_stored_bob_data(algo_hash: str) -> bytes | None:
    try:
        return bob_cache[algo_hash]
    except KeyError:
        filepath = configpaths.get('BOB') / algo_hash
        if filepath.exists():
            with open(str(filepath), 'r+b') as file:
                data = file.read()
            return data
    return None


def register_task(self: Any, task: Task) -> None:
    _tasks[id(self)].append(task)


def remove_task(task: Task, id_: int) -> None:
    try:
        _tasks[id_].remove(task)
    except Exception:
        pass
    else:
        if not _tasks[id_]:
            del _tasks[id_]


def cancel_tasks(obj: Any) -> None:
    id_ = id(obj)
    if id_ not in _tasks:
        return

    task_list = _tasks[id_]
    for task in task_list:
        task.cancel()


def check_finalize(obj: Any) -> None:
    if 'GAJIM_LEAK' not in os.environ:
        return

    name = obj.__class__.__name__
    logger = logging.getLogger('gajim.leak')
    finalizer = weakref.finalize(obj, logger.info, f'{name} has been finalized')

    g_objects: list[str] = []
    if isinstance(obj, GObject.Object):
        g_objects.append(name)

        def g_object_finalized():
            g_objects.remove(name)

        obj.weak_ref(g_object_finalized)

    def is_finalizer_ref(ref):
        try:
            return isinstance(ref[2][0], str)
        except Exception:
            return False

    def check_finalized():
        gc.collect()
        gc.collect()

        if g_objects:
            logger.warning('GObject not finalized: %s', name)

        tup = finalizer.peek()
        if tup is None:
            return

        logger.warning('%s not finalized', name)
        logger.warning('References:')
        for ref in gc.get_referrers(tup[0]):
            if is_finalizer_ref(ref):
                continue
            if isinstance(ref, dict):
                logger.warning('\n%s', pprint.pformat(ref))
            else:
                logger.warning(ref)

    GLib.timeout_add_seconds(2, check_finalized)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging
import math
from statistics import mean

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Graphene
from gi.repository import Gsk
from gi.repository import Gtk

from gajim.gtk.audio_player import AudioSampleT

log = logging.getLogger("gajim.gtk.preview_audio_visualizer")


class AudioVisualizerWidget(Gtk.Widget):
    def __init__(
        self,
        width: int = 340,
        height: int = 50,
        x_offset: int = 0,
    ) -> None:
        Gtk.Widget.__init__(
            self,
            margin_start=x_offset,
            width_request=width,
            height_request=height,
            halign=Gtk.Align.START,
        )

        self._peak_width = 2  # in px
        self._gap_height = 0.25
        self._width = width
        self._height = height
        self._is_LTR = bool(self.get_direction() == Gtk.TextDirection.LTR)

        self._samples: AudioSampleT = []
        self._seek_position = -1.0
        self._position = 0.0
        self._animation_index = 0
        self._animation_period = 1

        self._waveform_path = None

        style_manager = Adw.StyleManager.get_default()
        accent_color = style_manager.get_accent_color_rgba()

        self._color_progress = accent_color

        color_seek = Gdk.RGBA(
            red=accent_color.red * 1.3,
            green=accent_color.green * 1.3,
            blue=accent_color.blue * 1.3,
            alpha=accent_color.alpha,
        )
        self._color_seek = color_seek

        color_default = accent_color.copy()
        color_default.alpha = color_default.alpha - 0.4
        self._color_default = color_default

    def get_effective_width(self):
        return self._width

    def set_parameters(self, position: float, animation_period: int = 1) -> None:
        self._position = position
        self._animation_period = animation_period

    def set_samples(self, samples: AudioSampleT) -> None:
        if self._is_static():
            samples = self._average_samples(samples)
            samples = self._normalize_samples(samples)
        self._samples = self._rescale_samples(samples)
        self._waveform_path = self._create_waveform_path()

    def render_animated_graph(self, animation_index: int = 0) -> None:
        if not self._samples:
            log.debug("Render animated graph: No samples")
            return

        self._animation_index = animation_index
        self._waveform_path = self._create_waveform_path()
        self.queue_draw()

    def render_static_graph(self, position: float, seek_position: float = -1.0) -> None:
        if not self._samples:
            return

        self._position = position
        self._seek_position = seek_position
        self.queue_draw()

    def do_snapshot(self, snapshot: Gtk.Snapshot) -> None:
        if self._waveform_path is None:
            log.debug("Waveform Path is None")
            return

        if not self._is_LTR:
            # rotate 180° around the center
            snapshot.translate(Graphene.Point().init(self._width / 2, self._height / 2))
            snapshot.rotate(180)
            snapshot.translate(
                Graphene.Point().init(-self._width / 2, -self._height / 2)
            )

        play_pos = self._pixel_pos(self._position)

        # Default
        snapshot.push_clip(Graphene.Rect().init(0, 0, self._width, self._height))
        snapshot.append_fill(
            self._waveform_path, Gsk.FillRule.WINDING, self._color_default
        )
        snapshot.pop()

        # Progress
        snapshot.push_clip(Graphene.Rect().init(0, 0, play_pos, self._height))
        snapshot.append_fill(
            self._waveform_path, Gsk.FillRule.WINDING, self._color_progress
        )
        snapshot.pop()

        # Seek
        if self._seek_position >= 0:
            seek_pos = self._pixel_pos(self._seek_position)
            start_seek = min(play_pos, seek_pos)
            end_seek = max(play_pos, seek_pos)
            width_seek = end_seek - start_seek

            snapshot.push_clip(
                Graphene.Rect().init(start_seek, 0, width_seek, self._height)
            )
            snapshot.append_fill(
                self._waveform_path, Gsk.FillRule.WINDING, self._color_seek
            )
            snapshot.pop()

    def _is_static(self):
        return self._animation_period == 1

    def _average_samples(self, samples: AudioSampleT) -> AudioSampleT:
        # Create a subset by taking the average of three samples
        # around every nth sample
        num_divisions = int(self._width / (self._peak_width * 2))

        if num_divisions == 0:
            log.error("Number of divisions is zero")
            return []

        n = math.floor(len(samples) / num_divisions) + 1

        if n < 2:
            return samples

        samples_averaged: AudioSampleT = []
        samples1, samples2 = zip(*samples, strict=True)
        for i in range(2, int(len(samples) - n / 2), n):
            index = int(i + n / 2)
            avg1 = mean(samples1[index - 1 : index + 1])
            avg2 = mean(samples2[index - 1 : index + 1])
            samples_averaged.append((avg1, avg2))
        return samples_averaged

    def _normalize_samples(self, samples: AudioSampleT) -> AudioSampleT:
        if not samples:
            log.error("No samples")
            return []

        # Normalize both channels using the same scale
        max_elem = max(max(samples))  # noqa: PLW3301
        min_elem = min(min(samples))  # noqa: PLW3301
        delta = max_elem - min_elem
        if delta > 0:
            samples_normalized = [
                ((val1 - min_elem) / delta, (val2 - min_elem) / delta)
                for val1, val2 in samples
            ]
            return samples_normalized
        else:
            return samples

    def _rescale_samples(self, samples: AudioSampleT):
        """
        Recorded volume is perceived louder than the visual pretends.
        Therefore, rescale lower peaks bit. The sin(sqrt) function is
        a bit steeper than a shifted and normalized log function.
        """
        samples_rescaled = [
            (
                math.sin(math.sqrt(math.fabs(val1))) / math.sin(1),
                math.sin(math.sqrt(math.fabs(val2))) / math.sin(1),
            )
            for val1, val2 in samples
        ]
        return samples_rescaled

    def _pixel_pos(self, pos: float) -> float:
        return pos * self._width

    def _create_waveform_path(self) -> Gsk.Path | None:
        if not self._samples:
            return None

        peak_width = self._peak_width

        # determines the spacing between the amplitudes
        x_shift = (self._width - 2 * peak_width) / len(self._samples)
        x_shift_anime = x_shift * self._animation_index / self._animation_period
        x = x_shift - x_shift_anime

        path_builder = Gsk.PathBuilder.new()
        for i in range(len(self._samples)):
            sample1 = self._samples[i][0]
            sample2 = self._samples[i][1]

            rounded_rec = self._rounded_rec(
                x,
                self._height / 2,
                peak_width,
                sample1,
                sample2,
            )
            path_builder.add_path(rounded_rec)
            x += x_shift

        return path_builder.to_path()

    def _rounded_rec(
        self,
        x: float,
        y: float,
        width: float,
        height1: float,
        height2: float,
    ) -> Gsk.Path:
        # Don't insert a gap if bars are too small
        if height1 + height2 < 3 or not self._is_static():
            gap = 0.0
        else:
            gap = self._gap_height  # between upper and lower peak in pixels

        radius = 1  # radius of the arcs on top and bottom of the amplitudes
        scaling = self._height / 2 - radius - gap / 2  # peak scaling factor

        # Set a minimum height to improve visibility for quasi silence
        height1 = max(0.5, height1 * scaling)
        height2 = max(0.5, height2 * scaling)

        # Draws a rectangle of width w and total height of h1+h2
        # The top and bottom edges are curved to the outside
        # A --- B --- C
        # |           |
        # F --- E --- D

        # Up
        # A -- B -- C
        pb = Gsk.PathBuilder.new()

        pb.move_to(x, y + height1)
        pb.conic_to(x + width / 2, y + height1 + width / 2, x + width, y + height1, 30)
        # C -- D
        pb.line_to(x + width, y + gap)
        # D -- F
        pb.line_to(x, y + gap)
        # F -- A
        pb.line_to(x, y + height1)
        pb.close()

        # Down
        # C -- D
        pb.move_to(x + width, y - gap)
        pb.line_to(x + width, y - height2)
        # D -- E -- F
        pb.conic_to(x + width / 2, y - height2 - width / 2, x, y - height2, 30)
        # F -- A
        pb.line_to(x, y - gap)
        # A -- C
        pb.line_to(x + width, y - gap)
        pb.close()

        return pb.to_path()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Literal
from typing import overload

import logging

from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango
from nbxmpp.const import AdHocAction
from nbxmpp.const import AdHocNoteType
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules import dataforms
from nbxmpp.modules.dataforms import MultipleDataForm
from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.simplexml import Node
from nbxmpp.structs import AdHocCommand
from nbxmpp.structs import AdHocCommandNote
from nbxmpp.task import Task

from gajim.common import app
from gajim.common.helpers import to_user_string
from gajim.common.i18n import _
from gajim.common.util.text import process_non_spacing_marks

from gajim.gtk.assistant import Assistant
from gajim.gtk.assistant import AssistantErrorPage
from gajim.gtk.assistant import AssistantPage
from gajim.gtk.assistant import AssistantProgressPage
from gajim.gtk.assistant import DefaultPage
from gajim.gtk.dataform import DataFormWidget
from gajim.gtk.util.misc import container_remove_all
from gajim.gtk.util.misc import ensure_not_destroyed

log = logging.getLogger("gajim.gtk.adhoc")


class AdHocCommands(Assistant):
    def __init__(self, account: str, jids: list[str]) -> None:
        Assistant.__init__(self, width=600, height=500)
        self.account = account
        self.jid = str(jids[0])  # TODO: Add resource chooser

        self._destroyed = False

        self._client = app.get_client(account)

        self.add_button("cancel", _("Cancel"))
        self.add_button("prev", _("Previous"))
        self.add_button("next", _("Next"), complete=True, css_class="suggested-action")
        self.add_button(
            "complete", _("Complete"), complete=True, css_class="suggested-action"
        )
        self.add_button("commands", _("Commands"), css_class="suggested-action")
        self.add_button("execute", _("Execute"), css_class="suggested-action")

        self.add_pages(
            {
                "request": RequestCommandList(),
                "commands": Commands(),
                "stage": Stage(),
                "completed": Completed(),
                "error": Error(),
                "end": End(),
                "executing": Executing(),
            }
        )

        commands_page = self.get_page("commands")
        self._connect(commands_page, "execute", self._on_execute)

        self._connect(self, "button-clicked", self._on_button_clicked)

        self._client.get_module("AdHocCommands").request_command_list(
            self.jid, callback=self._received_command_list
        )

    @overload
    def get_page(self, name: Literal["request"]) -> RequestCommandList: ...

    @overload
    def get_page(self, name: Literal["commands"]) -> Commands: ...

    @overload
    def get_page(self, name: Literal["stage"]) -> Stage: ...

    @overload
    def get_page(self, name: Literal["completed"]) -> Completed: ...

    @overload
    def get_page(self, name: Literal["error"]) -> Error: ...

    @overload
    def get_page(self, name: Literal["end"]) -> End: ...

    @overload
    def get_page(self, name: Literal["executing"]) -> Executing: ...

    def get_page(self, name: str) -> AssistantPage:
        return self._pages[name]

    @ensure_not_destroyed
    def _received_command_list(self, task: Task) -> None:
        try:
            commands = cast(list[AdHocCommand] | None, task.finish())
        except StanzaError as error:
            if error.condition == "feature-not-implemented":
                commands = []
            else:
                self._set_error(to_user_string(error), False)
                return

        except MalformedStanzaError:
            self._set_error(_("Invalid server response"), False)
            return

        if not commands:
            self._set_end(_("No commands available"))
            return

        commands_page = self.get_page("commands")
        commands_page.add_commands(commands)
        self.show_page("commands")

    @ensure_not_destroyed
    def _received_stage(self, task: Task) -> None:
        try:
            stage = cast(AdHocCommand, task.finish())
        except StanzaError as error:
            self._set_error(to_user_string(error), True)
            return

        except MalformedStanzaError:
            self._set_error(_("Invalid server response"), False)
            return

        page_name = "stage"
        if stage.is_completed:
            page_name = "completed"

        page = self.get_page(page_name)
        page.process_stage(stage)
        self.show_page(page_name)

    def _set_error(self, text: str, show_command_button: bool) -> None:
        error_page = self.get_page("error")
        error_page.set_show_commands_button(show_command_button)
        error_page.set_text(text)
        self.show_page("error")

    def _set_end(self, text: str = "") -> None:
        end = self.get_page("end")
        end.set_text(text)
        self.show_page("end")

    def _cleanup(self) -> None:
        Assistant._cleanup(self)
        self._destroyed = True

    def _on_button_clicked(self, _assistant: AdHocCommands, button_name: str) -> None:
        if button_name == "commands":
            self._client.get_module("AdHocCommands").request_command_list(
                self.jid, callback=self._received_command_list
            )

        elif button_name == "execute":
            self._on_execute()

        elif button_name in ("prev", "next", "complete"):
            self._on_stage_action(AdHocAction(button_name))

        elif button_name == "cancel":
            self._on_cancel()

        else:
            raise ValueError("Invalid button name: %s" % button_name)

    def _on_stage_action(self, action: AdHocAction) -> None:
        stage_page = self.get_page("stage")
        command, dataform = stage_page.stage_data
        if action == AdHocAction.PREV:
            dataform = None

        self._client.get_module("AdHocCommands").execute_command(
            command, action=action, dataform=dataform, callback=self._received_stage
        )

        self.show_page("executing")
        stage_page = self.get_page("stage")
        stage_page.clear()

    def _on_execute(self, *args: Any) -> None:
        commands_page = self.get_page("commands")
        command = commands_page.get_selected_command()
        if command is None:
            return

        self._client.get_module("AdHocCommands").execute_command(
            command, action=AdHocAction.EXECUTE, callback=self._received_stage
        )

        self.show_page("executing")

    def _on_cancel(self) -> None:
        stage_page = self.get_page("stage")
        command, _ = stage_page.stage_data
        self._client.get_module("AdHocCommands").execute_command(
            command, AdHocAction.CANCEL
        )
        self.show_page("commands")


class Commands(AssistantPage):
    __gsignals__ = {
        "execute": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self) -> None:
        AssistantPage.__init__(self)

        self.set_valign(Gtk.Align.FILL)
        self.complete = True
        self.title = _("Command List")

        self._commands: dict[str, AdHocCommand] = {}
        self._scrolled = Gtk.ScrolledWindow(
            max_content_width=400,
            max_content_height=400,
            propagate_natural_height=True,
            hscrollbar_policy=Gtk.PolicyType.NEVER,
            vscrollbar_policy=Gtk.PolicyType.AUTOMATIC,
            vexpand=True,
            has_frame=True,
        )

        self._treeview = Gtk.TreeView()
        self._treeview.add_css_class("gajim-treeview")
        self._store = Gtk.ListStore(str, str)
        self._treeview.set_model(self._store)
        column = Gtk.TreeViewColumn(_("Commands"), Gtk.CellRendererText(), text=0)
        column.set_expand(True)
        self._treeview.append_column(column)

        self._connect(self._treeview, "row-activated", self._on_row_activate)
        self._treeview.set_search_equal_func(self._search_func)

        self._scrolled.set_child(self._treeview)
        self.append(self._scrolled)

    @staticmethod
    def _search_func(
        model: Gtk.TreeModel, _column: int, search_text: str, iter_: Gtk.TreeIter
    ) -> bool:
        return search_text.lower() not in model[iter_][0].lower()

    def _on_row_activate(
        self, _tree_view: Gtk.TreeView, _path: Gtk.TreePath, _column: Gtk.TreeViewColumn
    ) -> None:
        self.emit("execute")

    def add_commands(self, commands: list[AdHocCommand]) -> None:
        self._store.clear()
        self._commands = {}
        for command in commands:
            key = f"{command.jid}:{command.node}"
            self._commands[key] = command
            self._store.append((command.name, key))

    def get_selected_command(self) -> AdHocCommand | None:
        model, treeiter = self._treeview.get_selection().get_selected()
        if treeiter is None:
            return None
        key = model[treeiter][1]
        return self._commands[key]

    def get_visible_buttons(self) -> list[str]:
        return ["execute"]


class Stage(AssistantPage):
    def __init__(self) -> None:
        AssistantPage.__init__(self)

        self.set_valign(Gtk.Align.FILL)
        self.complete = False
        self.title = _("Stage")

        self._dataform_widget = None
        self._notes: list[Gtk.Label] = []
        self._last_stage_data: AdHocCommand | None = None
        self.default = None

    @property
    def stage_data(
        self,
    ) -> tuple[AdHocCommand, SimpleDataForm | MultipleDataForm | None]:
        assert self._last_stage_data is not None
        form = None
        if self._dataform_widget is not None:
            form = self._dataform_widget.get_submit_form()
        assert self._dataform_widget is not None
        return self._last_stage_data, form

    @property
    def actions(self) -> set[AdHocAction] | None:
        assert self._last_stage_data is not None
        return self._last_stage_data.actions

    def clear(self) -> None:
        self._show_form(None)
        self._show_notes(None)
        self._last_stage_data = None

    def process_stage(self, stage_data: AdHocCommand) -> None:
        self._last_stage_data = stage_data
        self._show_notes(stage_data.notes)
        self._show_form(stage_data.data)
        self.default = stage_data.default

    def _show_form(self, form: Node | None) -> None:
        if self._dataform_widget is not None:
            self.remove(self._dataform_widget)
        if form is None:
            return
        form = dataforms.extend_form(node=form)
        options = {"entry-activates-default": True}
        self._dataform_widget = DataFormWidget(form, options)
        self._connect(self._dataform_widget, "is-valid", self._on_is_valid)
        self._dataform_widget.validate()
        self.append(self._dataform_widget)

    def _show_notes(self, notes: list[AdHocCommandNote] | None) -> None:
        for note in self._notes:
            self.remove(note)
        self._notes = []

        if notes is None:
            return

        for note in notes:
            label = Gtk.Label(
                label=process_non_spacing_marks(note.text),
                wrap=True,
                wrap_mode=Pango.WrapMode.WORD_CHAR,
            )
            label.set_visible(True)
            self._notes.append(label)
            self.append(label)

    def _on_is_valid(self, _widget: DataFormWidget, is_valid: bool) -> None:
        self.complete = is_valid
        self.update_page_complete()

    def get_visible_buttons(self) -> list[str]:
        assert self._last_stage_data is not None
        assert self._last_stage_data.actions is not None
        return [action.value for action in self._last_stage_data.actions]

    def get_default_button(self) -> str:
        assert self._last_stage_data is not None
        assert self._last_stage_data.default is not None
        return self._last_stage_data.default.value


class Completed(AssistantPage):
    def __init__(self) -> None:
        AssistantPage.__init__(self)

        self.set_valign(Gtk.Align.FILL)
        self.complete = True
        self.title = _("Completed")
        self._severity = AdHocNoteType.INFO

        self._dataform_widget = None

        self._icon = SeverityIcon(self._severity)
        self._icon.set_visible(True)

        self._label = Gtk.Label(label=_("Completed"))
        self._label.set_visible(True)

        self._icon_text = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self._icon_text.set_spacing(12)
        self._icon_text.set_halign(Gtk.Align.CENTER)
        self._icon_text.append(self._icon)
        self._icon_text.append(self._label)
        self.append(self._icon_text)

        self._notes = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        self.append(self._notes)

    def process_stage(self, stage_data: AdHocCommand) -> None:
        self._reset_severity()
        self._show_notes(stage_data.notes)
        self._show_form(stage_data.data)
        self._show_icon_text(stage_data.data is None)
        self._show_icon(len(stage_data.notes or []) <= 1)

    def _set_status(self, status: str):
        self.title = status
        self._label.set_label(status)

    def _show_icon_text(self, show: bool) -> None:
        if show:
            self.set_valign(Gtk.Align.CENTER)
            self._icon_text.set_visible(True)
        else:
            self.set_valign(Gtk.Align.FILL)
            self._icon_text.set_visible(False)

        if self._severity == AdHocNoteType.INFO:
            self._set_status(_("Completed"))
        elif self._severity == AdHocNoteType.WARN:
            self._set_status(_("Warning"))
        elif self._severity == AdHocNoteType.ERROR:
            self._set_status(_("Error"))

    def _show_form(self, form: Node | None) -> None:
        if self._dataform_widget is not None:
            self.remove(self._dataform_widget)
        if form is None:
            return

        form = dataforms.extend_form(node=form)

        self._dataform_widget = DataFormWidget(form, options={"read-only": True})
        self.append(self._dataform_widget)

    def _show_notes(self, notes: list[AdHocCommandNote] | None) -> None:
        container_remove_all(self._notes)

        if notes is None:
            return

        for note in notes:
            box = Gtk.Box(margin_start=24, spacing=24)
            if len(notes) > 1:
                box.append(SeverityIcon(note.type))
            else:
                box.set_halign(Gtk.Align.CENTER)

            label = Gtk.Label(
                label=process_non_spacing_marks(note.text),
                natural_wrap_mode=Gtk.NaturalWrapMode.WORD,
                selectable=True,
                wrap=True,
                wrap_mode=Pango.WrapMode.WORD_CHAR,
            )
            box.append(label)

            self._notes.append(box)
            self._bump_severity(note.type)

    def _show_icon(self, show: bool):
        if show:
            self._icon.set_severity(self._severity)
        else:
            self._icon.set_visible(False)

    def _reset_severity(self):
        self._severity = AdHocNoteType.INFO

    def _bump_severity(self, severity: AdHocNoteType):
        if (
            severity == AdHocNoteType.WARN and self._severity != AdHocNoteType.ERROR
        ) or severity == AdHocNoteType.ERROR:
            self._severity = severity

    def get_visible_buttons(self) -> list[str]:
        return ["commands"]


class Error(AssistantErrorPage):
    def __init__(self) -> None:
        AssistantErrorPage.__init__(self)

        self._show_commands_button = False
        self.set_heading(_("An error occurred"))

    def set_show_commands_button(self, value: bool) -> None:
        self._show_commands_button = value

    def get_visible_buttons(self) -> list[str]:
        if self._show_commands_button:
            return ["commands"]
        return []


class End(DefaultPage):
    def __init__(self) -> None:
        DefaultPage.__init__(self)
        self.title = _("Completed")


class Executing(AssistantProgressPage):
    def __init__(self) -> None:
        AssistantProgressPage.__init__(self)
        self.set_title(_("Executing…"))
        self.set_text(_("Executing…"))


class RequestCommandList(AssistantProgressPage):
    def __init__(self) -> None:
        AssistantProgressPage.__init__(self)
        self.set_title(_("Requesting Command List"))
        self.set_text(_("Requesting Command List"))


class SeverityIcon(Gtk.Image):
    def __init__(self, severity: AdHocNoteType) -> None:
        Gtk.Image.__init__(self)
        self.set_severity(severity)

    def set_severity(self, severity: AdHocNoteType) -> None:
        self.remove_css_class("success")
        self.remove_css_class("warning")
        self.remove_css_class("error")
        if severity == AdHocNoteType.INFO:
            self.set_from_icon_name("lucide-check-symbolic")
            self.add_css_class("success")
        elif severity == AdHocNoteType.WARN:
            self.set_from_icon_name("lucide-circle-alert-symbolic")
            self.add_css_class("warning")
        elif severity == AdHocNoteType.ERROR:
            self.set_from_icon_name("lucide-circle-x-symbolic")
            self.add_css_class("error")



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

import locale
import logging
import time

from gi.repository import Adw
from gi.repository import Gtk
from omemo_dr.const import OMEMOTrust
from omemo_dr.structs import IdentityInfo

from gajim.common import app
from gajim.common import ged
from gajim.common import types
from gajim.common.events import AccountConnected
from gajim.common.events import AccountDisconnected
from gajim.common.ged import EventHelper
from gajim.common.helpers import generate_qr_code
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.omemo import compose_trust_uri

from gajim.gtk.alert import ConfirmationAlertDialog
from gajim.gtk.builder import get_builder
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import clear_listbox
from gajim.gtk.util.misc import remove_css_class
from gajim.gtk.util.window import open_window

log = logging.getLogger("gajim.gtk.omemo_trust_manager")


TRUST_DATA = {
    OMEMOTrust.UNTRUSTED: (
        "lucide-shield-x-symbolic",
        _("Untrusted"),
        "omemo-untrusted-color",
    ),
    OMEMOTrust.UNDECIDED: (
        "lucide-shield-question-mark-symbolic",
        _("Not Decided"),
        "omemo-undecided-color",
    ),
    OMEMOTrust.VERIFIED: (
        "lucide-shield-check-symbolic",
        _("Verified"),
        "omemo-verified-color",
    ),
    OMEMOTrust.BLIND: (
        "lucide-shield-symbolic",
        _("Blind Trust"),
        "omemo-blind-color",
    ),
}


class OMEMOTrustManager(Gtk.Box, EventHelper, SignalManager):
    def __init__(self, account: str, contact: types.ChatContactT | None = None) -> None:
        Gtk.Box.__init__(self)
        EventHelper.__init__(self)
        SignalManager.__init__(self)

        self._account = account
        self._contact = contact

        self._ui = get_builder("omemo_trust_manager.ui")
        self.append(self._ui.stack)

        self._connect(
            self._ui.manage_trust_button, "clicked", self._on_manage_trust_clicked
        )
        self._connect(
            self._ui.show_inactive_switch, "state-set", self._on_show_inactive
        )
        self._connect(
            self._ui.clear_devices_button, "clicked", self._on_clear_devices_clicked
        )
        self._connect(self._ui.copy_button, "clicked", self._on_copy_button_clicked)

        self._ui.list.set_sort_func(self._sort_func, None)

        self.register_events(
            [
                ("account-connected", ged.GUI2, self._on_account_state),
                ("account-disconnected", ged.GUI2, self._on_account_state),
            ]
        )

        if not app.account_is_connected(account):
            self._ui.stack.set_visible_child_name("no-connection")
            return

        self._update()

    def do_unroot(self) -> None:
        self.unregister_events()
        self._disconnect_all()
        self._ui.list.set_filter_func(None)
        self._ui.list.set_sort_func(None)

        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    def _update(self) -> None:
        client = app.get_client(self._account)
        if self._contact is None:
            self._contact = client.get_module("Contacts").get_contact(
                client.get_own_jid().bare
            )

        if isinstance(self._contact, BareContact) and self._contact.is_self:
            header_text = _("Other devices connected with your account")
            popover_qr_text = _(
                "Compare this code with the one shown on your "
                "contact’s screen to ensure the safety of "
                "your end-to-end encrypted chat."
            )

        else:
            assert isinstance(
                self._contact, BareContact | GroupchatContact | GroupchatParticipant
            )
            if self._contact.is_groupchat:
                header_text = (
                    _('Undecided fingerprints of devices connected with "%s"')
                    % self._contact.name
                )
            else:
                header_text = (
                    _('Fingerprints of devices connected with "%s"')
                    % self._contact.name
                )
            popover_qr_text = (
                _(
                    "Compare this code with the one shown on your "
                    "contact’s screen to ensure the safety of "
                    "your end-to-end encrypted chat "
                    "with %s."
                )
                % self._contact.name
            )

        self._ui.list_heading.set_text(header_text)
        self._ui.comparing_instructions.set_text(popover_qr_text)

        self._our_fpr_formatted = client.get_module(
            "OMEMO"
        ).backend.get_our_fingerprint(formatted=True)
        self._ui.our_fingerprint_row.set_title(self._our_fpr_formatted)
        self._ui.our_fingerprint_2.set_text(self._our_fpr_formatted)

        self.update()
        self._load_qrcode()

    def update(self) -> None:
        assert self._contact is not None
        clear_listbox(self._ui.list)

        if isinstance(self._contact, BareContact) and self._contact.is_self:
            self._ui.clear_devices_button.set_visible(True)
        else:
            self._ui.manage_trust_button.set_visible(True)
            if self._contact.is_groupchat:
                self._ui.list_heading_box.set_visible(False)

        assert isinstance(
            self._contact, BareContact | GroupchatContact | GroupchatParticipant
        )
        self._load_fingerprints(self._contact)

    def _on_account_state(self, event: AccountConnected | AccountDisconnected) -> None:
        if event.account != self._account:
            return

        if isinstance(event, AccountConnected):
            self._update()
            self._ui.stack.set_visible_child_name("manage-keys")
        else:
            self._ui.stack.set_visible_child_name("no-connection")

    @staticmethod
    def _sort_func(row1: DeviceRow, row2: DeviceRow, _user_data: Any) -> int:
        result = locale.strcoll(row1.address, row2.address)
        if result != 0:
            return result

        if row1.active != row2.active:
            return -1 if row1.active else 1

        if row1.trust != row2.trust:
            return -1 if row1.trust > row2.trust else 1
        return 0

    def _load_fingerprints(self, contact: types.ChatContactT) -> None:
        client = app.get_client(self._account)

        identity_info = client.get_module("OMEMO").backend.get_identity_infos(
            str(contact.jid),
            only_active=contact.is_groupchat,
            trust=OMEMOTrust.UNDECIDED if contact.is_groupchat else None,
        )

        for info in identity_info:
            self._ui.list.append(DeviceRow(contact, info))

        assert self._contact is not None
        if self._contact.is_groupchat:
            self._ui.list.set_visible(bool(identity_info))
            self._ui.undecided_placeholder.set_visible(not bool(identity_info))

    def _load_qrcode(self) -> None:
        client = app.get_client(self._account)
        uri = compose_trust_uri(
            client.get_own_jid(),
            [client.get_module("OMEMO").backend.get_our_identity()],
        )
        log.debug("Trust URI: %s", uri)
        self._ui.qr_code_image.set_from_paintable(generate_qr_code(uri))

    def _on_show_inactive(self, _switch: Gtk.Switch, _state: bool) -> None:
        self._ui.list.invalidate_filter()

    def _on_clear_devices_clicked(self, _button: Gtk.Button) -> None:
        def _on_response() -> None:
            client = app.get_client(self._account)
            client.get_module("OMEMO").clear_devicelist()

        ConfirmationAlertDialog(
            _("Clear Devices?"),
            _("This will clear the devices store for your account."),
            confirm_label=_("_Clear Devices"),
            callback=_on_response,
        )

    def _on_manage_trust_clicked(self, _button: Gtk.Button) -> None:
        assert self._contact is not None
        window = open_window("Preferences")
        window.show_page(f"{self._contact.account}-encryption-omemo")

    def _on_copy_button_clicked(self, _button: Gtk.Button) -> None:
        app.window.get_clipboard().set(self._our_fpr_formatted)


class DeviceRow(Adw.ActionRow, SignalManager):
    def __init__(
        self, contact: types.ChatContactT, identity_info: IdentityInfo
    ) -> None:
        Adw.ActionRow.__init__(self)
        SignalManager.__init__(self)
        self.add_css_class("omemo-device-row")

        self._contact = contact
        self._client = app.get_client(contact.account)

        self._address = identity_info.address
        self._identity_info = identity_info
        self._trust = identity_info.trust

        subtitle_entries: list[str] = []
        if contact.is_groupchat:
            subtitle_entries.append(self._address)

        self._formatted_fingerprint = self._identity_info.public_key.get_fingerprint(
            formatted=True
        )

        if self._identity_info.last_seen is not None and not contact.is_groupchat:
            last_seen_data = time.strftime(
                app.settings.get("date_time_format"),
                time.localtime(self._identity_info.last_seen),
            )
        else:
            last_seen_data = _("Never")
        subtitle_entries.append(_("Last seen: %s") % last_seen_data)

        if not self._identity_info.active:
            subtitle_entries.append(_("(inactive)"))

        self.set_title(self._formatted_fingerprint)
        self.set_subtitle(" ".join(subtitle_entries))

        self._trust_label = TrustLabel(identity_info.trust)
        self.add_suffix(self._trust_label)

        self._trust_button = TrustButton(self)
        self.add_suffix(self._trust_button)

        self._copy_button = Gtk.Button(
            icon_name="lucide-copy-symbolic",
            tooltip_text=_("Copy to Clipboard"),
            valign=Gtk.Align.CENTER,
            halign=Gtk.Align.END,
        )
        self._connect(self._copy_button, "clicked", self._on_copy_button_clicked)
        self.add_suffix(self._copy_button)

    def do_unroot(self) -> None:
        del self._trust_button
        self._disconnect_all()
        Adw.ActionRow.do_unroot(self)
        app.check_finalize(self)

    def delete_fingerprint(self, *args: Any) -> None:
        def _on_response() -> None:
            self._client.get_module("OMEMO").backend.delete_session(
                self._address, self._identity_info.device_id, delete_identity=True
            )

            listbox = cast(Gtk.ListBox, self.get_parent())
            listbox.remove(self)

        ConfirmationAlertDialog(
            _("Delete Fingerprint?"),
            _("Doing so will permanently delete this Fingerprint"),
            confirm_label=_("_Delete"),
            appearance="destructive",
            callback=_on_response,
        )

    def set_trust(self, trust: OMEMOTrust) -> None:
        self._trust = trust
        self._trust_label.set_trust(trust)

        self._client.get_module("OMEMO").backend.set_trust(
            self._address, self._identity_info.public_key, trust
        )

    @property
    def trust(self) -> OMEMOTrust:
        return self._trust

    @property
    def active(self) -> bool:
        return self._identity_info.active

    @property
    def address(self) -> str:
        return self._address

    def _on_copy_button_clicked(self, _button: Gtk.Button) -> None:
        app.window.get_clipboard().set(self._formatted_fingerprint)


class TrustLabel(Gtk.Box):
    def __init__(self, trust: OMEMOTrust) -> None:
        Gtk.Box.__init__(
            self,
            margin_end=12,
            spacing=6,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            halign=Gtk.Align.END,
        )

        self._label = Gtk.Label()
        self._image = Gtk.Image()

        self.append(self._image)
        self.append(self._label)

        self.set_trust(trust)

    def set_trust(self, trust: OMEMOTrust) -> None:
        remove_css_class(self._label, "omemo-")
        remove_css_class(self._image, "omemo-")

        icon_name, text, css_class = TRUST_DATA[trust]

        self._label.set_text(text)
        self._label.add_css_class(css_class)

        self._image.set_from_icon_name(icon_name)
        self._image.add_css_class(css_class)


class TrustButton(Gtk.MenuButton):
    def __init__(self, row: DeviceRow) -> None:
        Gtk.MenuButton.__init__(self, halign=Gtk.Align.END, valign=Gtk.Align.CENTER)
        self._row = row
        self._trust_popover = TrustPopver(row, self)

        image = Gtk.Image.new_from_icon_name("lucide-shield-symbolic")
        self.set_child(image)
        self.set_popover(self._trust_popover)
        self.set_tooltip_text(_("Set Trust"))

    def do_unroot(self) -> None:
        del self._trust_popover
        del self._row

        Gtk.MenuButton.do_unroot(self)
        app.check_finalize(self)


class TrustPopver(Gtk.Popover, SignalManager):
    def __init__(self, row: DeviceRow, trust_button: TrustButton) -> None:
        Gtk.Popover.__init__(self)
        SignalManager.__init__(self)
        self._row = row
        self._trust_button = trust_button

        self._listbox = Gtk.ListBox()
        self._listbox.set_selection_mode(Gtk.SelectionMode.NONE)
        self.update()
        self.set_child(self._listbox)
        self._connect(self._listbox, "row-activated", self._activated)
        self.add_css_class("menu")

    def do_unroot(self) -> None:
        Gtk.Popover.do_unroot(self)
        self._disconnect_all()
        del self._listbox
        del self._row
        del self._trust_button
        app.check_finalize(self)

    def _activated(self, _listbox: Gtk.ListBox, row: MenuOption) -> None:
        self.popdown()
        if row.type_ is None:
            self._row.delete_fingerprint()
        else:
            self._row.set_trust(row.type_)
            self.update()

    def update(self) -> None:
        self._listbox.remove_all()

        if self._row.trust != OMEMOTrust.VERIFIED:
            self._listbox.append(VerifiedOption())
        if self._row.trust != OMEMOTrust.BLIND:
            self._listbox.append(BlindOption())
        if self._row.trust != OMEMOTrust.UNTRUSTED:
            self._listbox.append(NotTrustedOption())
        self._listbox.append(DeleteOption())


class MenuOption(Gtk.ListBoxRow):
    def __init__(
        self, icon: str, label_text: str, color: str, type_: OMEMOTrust | None = None
    ) -> None:
        Gtk.ListBoxRow.__init__(self)

        self.type_ = type_
        self.icon = icon
        self.label = label_text
        self.color = color

        box = Gtk.Box()
        box.set_spacing(6)

        image = Gtk.Image.new_from_icon_name(icon)
        if color:
            image.add_css_class(color)

        label = Gtk.Label(label=label_text)

        box.append(image)
        box.append(label)
        self.set_child(box)


class BlindOption(MenuOption):
    def __init__(self) -> None:
        MenuOption.__init__(
            self,
            "lucide-shield-symbolic",
            _("Blind Trust"),
            "encrypted-color",
            OMEMOTrust.BLIND,
        )


class VerifiedOption(MenuOption):
    def __init__(self) -> None:
        MenuOption.__init__(
            self,
            "lucide-shield-check-symbolic",
            _("Verified"),
            "encrypted-color",
            OMEMOTrust.VERIFIED,
        )


class NotTrustedOption(MenuOption):
    def __init__(self) -> None:
        MenuOption.__init__(
            self,
            "lucide-circle-x-symbolic",
            _("Untrusted"),
            "error",
            OMEMOTrust.UNTRUSTED,
        )


class DeleteOption(MenuOption):
    def __init__(self) -> None:
        MenuOption.__init__(self, "lucide-trash-symbolic", _("Delete"), "")
